#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

TESTS = []

#  -------------------------------------------------------------------------
tag = 'grid 1D'
diag = True
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 1, 1)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5,), dtype='float64', order='F'))
g = CGL.newGridCoordinates(z, '{Grid#002}')
d = CGL.newDataArray(g, CGK.CoordinateR_s, NPY.ones((5,), dtype='float64', order='F'))
g = CGL.newGridCoordinates(z, '{Grid#003}')
d = CGL.newDataArray(g, CGK.CoordinateXi_s, NPY.ones((5,), dtype='float64', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'grid 2D'
diag = True
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 2, 2)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7), dtype='float64', order='F'))
g = CGL.newGridCoordinates(z, '{Grid#002}')
d = CGL.newDataArray(g, CGK.CoordinateXi_s, NPY.ones((5, 7), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateEta_s, NPY.ones((5, 7), dtype='float64', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'grid 3D #1'
diag = True
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'grid 3D #2'
diag = True
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 2, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7), dtype='float64', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'grid empty #1'
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
g = CGL.newGridCoordinates(z, '{Grid#002}')
g = CGL.newGridCoordinates(z, '{Grid#003}')
g = CGL.newGridCoordinates(z, '{Grid#004}')
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'grid empty #2'
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
r = CGL.newUserDefinedData(g, '{UserDefinedData}')
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'grid bad dims #1'
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, '{Base}', 3, 3)
z = CGL.newZone(b, '{Zone}', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
g = CGL.newGridCoordinates(z, 'GridCoordinates')
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((4, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------



from __future__ import print_function
import CGNS.MAP              as CGM
import CGNS.PAT.cgnskeywords as CGK
import CGNS.PAT.cgnsutils    as CGU
import CGNS.PAT.cgnslib      as CGL

import CGNS.APP.probe.arrayutils as ARU


def parseTree(filename):
    flags = CGM.S2P_DEFAULT
    (tree, l) = CGM.load(filename, flags, 0, None, [], None)
    typepath = [CGK.CGNSTree_ts, CGK.CGNSBase_ts, CGK.Zone_ts, CGK.Elements_ts]
    elist = CGU.getAllNodesByTypeList(tree, typepath)
    sn = 0
    sl = []
    sp = ARU.SectionParse()
    mr = 1
    for e in elist:
        print('Parse ', e)
        sn += 1
        ne = CGU.getNodeByPath(tree, e)[1]
        et = ne[0]
        eb = ne[1]
        ea = CGU.getNodeByPath(tree, e + '/' + CGK.ElementConnectivity_s)[1]
        if (et in sp.QUAD_SURFACE):
            sl.append(sp.extQuadFacesPoints(ea, et, sn, mr, eb))
        if (et in sp.TRI_SURFACE):
            sl.append(sp.extTriFacesPoints(ea, et, sn, mr, eb))
        mr = sl[-1][-1]

# for s in sl:
#       print CGK.ElementType_l[s[0]]
#       if (s[0]==CGK.QUAD_4):
#         for p in range(len(s[1])/4):
#             print s[1][4*p+0],s[1][4*p+1],s[1][4*p+2],s[1][4*p+3]

# --- last line



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  ---------------------------------------------------------------------------
#
from __future__ import unicode_literals
#
import CGNS.PAT.cgnslib as C
import CGNS.PAT.cgnskeywords as K
import CGNS.PAT.cgnserrors as E
import numpy as N
#
#
data = C.newArbitraryGridMotion(None, '{ArbitraryGridMotion}')
C.newRind(data, N.array([0, 0, 0, 0, 1, 1]))
C.newGridLocation(data)
C.newDataArray(data, K.GridVelocityX_s)
C.newDataArray(data, K.GridVelocityY_s)
C.newDataArray(data, K.GridVelocityZ_s)
C.newDataClass(data)
C.newDimensionalUnits(data)
C.newUserDefinedData(data, '{UserDefinedData}')
C.newDescriptor(data, '{Descriptor}')
#
status = '11.3'
comment = 'Full SIDS with all optionals'
pattern = [data, status, comment]
#



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

from CGNS.NAV.moption import Q7OptionContext as OCTXT

import sys
import string

from qtpy.QtCore import Qt
#from qtpy.QtWidgets import QWidget

from CGNS.NAV.Q7InfoWindow import Ui_Q7InfoWindow
from CGNS.NAV.wfingerprint import Q7Window


# -----------------------------------------------------------------
class Q7Info(Q7Window, Ui_Q7InfoWindow):
    def __init__(self, parent, data, fgprint):
        Q7Window.__init__(self, Q7Window.VIEW_INFO, parent, None, None)
        self.bClose.clicked.connect(self.reject)
        self.bInfo.clicked.connect(self.infoInfoView)
        self._fgprint = fgprint
        self._data = data

    def infoInfoView(self):
        self._control.helpWindow('Info')

    def show(self):
        self.reset()
        self.bHasBeenModified.setVisible(False)
        if self._fgprint.fileHasChanged():
            self.bHasBeenModified.setVisible(True)
        super(Q7Info, self).show()

    def reset(self):
        for k in dir(self):
            if k in self._data:
                v = self._data[k]
                if isinstance(v, bool):
                    getattr(self, k).setEnabled(True)
                    if v:
                        getattr(self, k).setCheckState(Qt.Checked)
                    else:
                        getattr(self, k).setCheckState(Qt.Unchecked)
                    getattr(self, k).setEnabled(False)
                if isinstance(v, (str, bytes, int, float)):
                    getattr(self, k).setText(str(v))
                    # getattr(self,k).setFont(OCTXT.FixedFontTable)

    def reject(self):
        self.close()

# -----------------------------------------------------------------



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  ---------------------------------------------------------------------------
#
from __future__ import unicode_literals
import CGNS.PAT.cgnslib      as C
import CGNS.PAT.cgnserrors   as E
import CGNS.PAT.cgnskeywords as K
import numpy             as N

data = ['TurbulenceClosure', None, [], 'TurbulenceClosure_t']
status = ''
comment = ''
pattern = [data, status, comment]



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
# TESTING VAL
#
from __future__ import unicode_literals
from __future__ import print_function
import unittest

import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

import importlib
import os
import string
import subprocess


class VALTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_001_Base_write_close(self):
        import CGNS.VAL.suite.run
        CGNS.VAL.suite.run.runall()


# ---
print('-' * 70 + '\nCGNS.VAL test suite')
suite = unittest.TestLoader().loadTestsFromTestCase(VALTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

import sys
import gc
import functools

import CGNS.PAT.cgnskeywords as CGK
import CGNS.PAT.cgnsutils    as CGU
import CGNS.PAT.SIDS         as CGS

from qtpy.QtCore import Qt, QModelIndex
from qtpy.QtWidgets import (QStyledItemDelegate, QLineEdit, QComboBox, QSizePolicy,
                            QStyleOptionViewItem, QMenu, QAction)
from qtpy.QtGui import (QFont, QPalette, QScreen, QKeySequence)

from CGNS.pyCGNSconfig import HAS_VTK

from CGNS.NAV.Q7TreeWindow import Ui_Q7TreeWindow
from CGNS.NAV.wform import Q7Form
from CGNS.NAV.wpattern import Q7PatternList
from CGNS.NAV.wquery import Q7Query, Q7SelectionList
from CGNS.NAV.wdiag import Q7CheckList
from CGNS.NAV.wlink import Q7LinkList
from CGNS.NAV.mtree import (Q7TreeModel, Q7TreeItem)
from CGNS.NAV.wfingerprint import Q7Window, Q7FingerPrint
from CGNS.NAV.moption import Q7OptionContext as OCTXT

import CGNS.NAV.mtree as NMT
import CGNS.NAV.wmessages as MSG

(CELLCOMBO, CELLTEXT) = range(2)
CELLEDITMODE = (CELLCOMBO, CELLTEXT)


# -----------------------------------------------------------------
class Q7TreeItemDelegate(QStyledItemDelegate):
    def __init__(self, wtree, model):
        QStyledItemDelegate.__init__(self, wtree)
        self._parent = wtree
        self._mode = CELLTEXT
        self._model = model

    def createEditor(self, parent, option, index):
        if (self._parent.isLocked()): return None
        if (index.internalPointer().sidsIsCGNSTree()): return None
        if (index.internalPointer().sidsIsLink()): return None
        if (index.internalPointer().sidsIsLinkChild()): return None
        ws = option.rect.width()
        hs = option.rect.height() + 4
        xs = option.rect.x()
        ys = option.rect.y() - 2
        if (not index.internalPointer().hasEditCheck() or
                (index.column() in [NMT.COLUMN_NAME])):
            self._mode = CELLTEXT
            editor = QLineEdit(parent)
            editor.transgeometry = (xs, ys, ws, hs)
            editor.installEventFilter(self)
            self.setEditorData(editor, index)
            return editor
        if index.column() in [NMT.COLUMN_VALUE]:
            node = self._parent.modelData(index)
            if node.hasValueView():
                pt = node.sidsPath().split('/')[1:]
                lt = node.sidsTypePath()
                fc = self._parent._control._control.userFunctionFromPath(pt, lt)
                if (fc is not None):
                    en = fc.getEnumerate(pt, lt)
                else:
                    en = node.sidsValueEnum()
                if (en is None):
                    self._mode = CELLTEXT
                    editor = QLineEdit(parent)
                    editor.transgeometry = (xs, ys, ws, hs)
                else:
                    self._mode = CELLCOMBO
                    editor = QComboBox(parent)
                    editor.transgeometry = (xs, ys, ws, hs)
                    editor.addItems(en)
                    try:
                        tix = en.index(node.sidsValue().tostring().decode('ascii'))
                    except ValueError:
                        editor.insertItem(0, node.sidsValue().tostring().decode('ascii'))
                        tix = 0
                    editor.setCurrentIndex(tix)
                editor.installEventFilter(self)
                self.setEditorData(editor, index)
                return editor
        if (index.column() == NMT.COLUMN_SIDS):
            self._mode = CELLCOMBO
            editor = QComboBox(parent)
            editor.transgeometry = (xs, ys, ws, hs)
            tnode = self._parent.modelData(index)
            itemslist = tnode.sidsTypeList()
            editor.addItems(itemslist)
            try:
                tix = itemslist.index(tnode.sidsType())
            except ValueError:
                editor.insertItem(0, tnode.sidsType())
                tix = 0
            editor.setCurrentIndex(tix)
            editor.installEventFilter(self)
            self.setEditorData(editor, index)
            return editor
        if (index.column() == NMT.COLUMN_DATATYPE):
            self._mode = CELLCOMBO
            editor = QComboBox(parent)
            editor.transgeometry = (xs, ys, ws, hs)
            editor.setProperty("Q7SIDSDataTypeComboBox", "True")
            itemslist = self._parent.modelData(index).sidsDataType(all=True)
            editor.addItems(itemslist)
            editor.installEventFilter(self)
            sizePolicy = QSizePolicy(QSizePolicy.Minimum,QSizePolicy.Minimum)
            editor.setSizePolicy(sizePolicy)
            self.setEditorData(editor, index)
            return editor
        return None

    def setEditorData(self, editor, index):
        if (self._mode == CELLTEXT):
            value = index.data()
            editor.clear()
            editor.insert(value)
        elif (self._mode == CELLCOMBO):
            value = index.data()
            ix = editor.findText(value)
            print('VALUE',value,ix)
            if (ix != -1): editor.setCurrentIndex(ix)
        else:
            pass

    def setModelData(self, editor, model, index):
        value = None
        if (self._mode == CELLCOMBO):
            value = editor.currentText()
        if (self._mode == CELLTEXT):
            value = editor.text()
        pth = self._parent.modelData(index).sidsPath()
        model.setData(index, value, role=Qt.EditRole)
        node = index.internalPointer().lastEdited()
        if (node is not None):
            nindex = self._parent.model().indexByPath(node.sidsPath())
            self._parent.setLastEntered(nindex)
        if (self._parent._control._selectwindow is not None):
            self._parent._control._selectwindow.reset()

    def updateEditorGeometry(self, editor, option, index):
        editor.setGeometry(*editor.transgeometry)

    def paint(self, painter, option, index):
        if (self._parent.modelIndex(index).column() == NMT.COLUMN_NAME):
            if (self._parent.modelData(index).sidsName()
                not in OCTXT._ReservedNames):
                option.font.setWeight(QFont.Bold)
            uf = self._parent.modelData(index).userState()
            if (uf in NMT.USERSTATES):
                if (self._model.hasUserColor(uf)):
                    cl = self._model.getUserColor(uf)
                    option.palette.brush(QPalette.Text).setColor(cl)
            QStyledItemDelegate.paint(self, painter, option, index)
            option.font.setWeight(QFont.Light)
        elif (index.column() in [NMT.COLUMN_VALUE, NMT.COLUMN_DATATYPE]):
            option.font.setFamily(OCTXT.Table_Family)
            if (index.column() == NMT.COLUMN_DATATYPE):
                # option.font.setPointSize(8)
                pass
            QStyledItemDelegate.paint(self, painter, option, index)
        elif (index.column() in NMT.COLUMN_FLAGS):
            option.decorationPosition = QStyleOptionViewItem.Top
            QStyledItemDelegate.paint(self, painter, option, index)
            option.decorationPosition = QStyleOptionViewItem.Left
        else:
            QStyledItemDelegate.paint(self, painter, option, index)

    def doRelease(self):
        self._model = None


# -----------------------------------------------------------------
class Q7Tree(Q7Window, Ui_Q7TreeWindow):
    def __init__(self, control, path, fgprintindex):
        Q7Window.__init__(self, Q7Window.VIEW_TREE, control, path, fgprintindex)
        self._depthExpanded = 0
        self._lastEntered = None
        self.lastdiag = None
        self._linkwindow = None
        self._querywindow = None
        self._vtkwindow = None
        self._selectwindow = None
        self._column = {NMT.COLUMN_SIDS: OCTXT.ShowSIDSColumn,
                        NMT.COLUMN_FLAG_LINK: OCTXT.ShowLinkColumn,
                        NMT.COLUMN_FLAG_SELECT: OCTXT.ShowSelectColumn,
                        NMT.COLUMN_FLAG_CHECK: OCTXT.ShowCheckColumn,
                        NMT.COLUMN_FLAG_USER: OCTXT.ShowUserColumn,
                        NMT.COLUMN_SHAPE: OCTXT.ShowShapeColumn,
                        NMT.COLUMN_DATATYPE: OCTXT.ShowDataTypeColumn}
        self.selectForLinkSrc = None  # one link source per tree view allowed

        #self.treeview.expanded[QModelIndex].connect(self.expandNode)
        self.treeview.collapsed.connect(self.collapseNode)
        self.treeview.pressed[QModelIndex].connect(self.clickedPressedNode)
        self.treeview.customContextMenuRequested.connect(self.clickedNode)

        # QObject.connect(self.treeview,
        #                SIGNAL("expanded(QModelIndex)"),
        #                self.expandNode)
        # QObject.connect(self.treeview,
        #                SIGNAL("collapsed()"),
        #                self.collapseNode)
        # QObject.connect(self.treeview,
        #                SIGNAL("pressed(QModelIndex)"),
        #                self.clickedPressedNode)
        # QObject.connect(self.treeview,
        #                SIGNAL("customContextMenuRequested(QPoint)"),
        #                self.clickedNode)

        self.bSave.clicked.connect(self.savetree)
        self.lockable(self.bSave)
        self.bQueryView.clicked.connect(self.queryview)
        self.lockable(self.bQueryView)
        self.bSaveAs.clicked.connect(self.savetreeas)
        self.lockable(self.bSaveAs)
        self.bInfo.clicked.connect(self.infoTreeView)
        self.bZoomIn.clicked.connect(self.expandLevel)
        self.bZoomOut.clicked.connect(self.collapseLevel)
        self.bZoomAll.clicked.connect(self.expandMinMax)
        self.bFormView.clicked.connect(self.formview)
        self.bMarkAll.clicked.connect(self.markall)
        self.bUnmarkAll_1.clicked.connect(self.unmarkall)
        self.bUnmarkAll_2.clicked.connect(self.unmarkall)
        self.bPreviousMark.clicked.connect(self.previousmark)
        self.bNextMark.clicked.connect(self.nextmark)
        self.bSwapMarks.clicked.connect(self.swapmarks)
        self.bMarksAsList.clicked.connect(self.selectionlist)
        self.bVTKView.clicked.connect(self.vtkview)
        self.lockable(self.bVTKView)
        self.bScreenShot.clicked.connect(self.screenshot)
        self.bCheck.clicked.connect(self.check)
        self.bCheckList.clicked.connect(self.checklist)
        self.bClearChecks.clicked.connect(self.clearchecks)
        self.bLinkView.clicked.connect(self.linklist)
        self.bPatternView.clicked.connect(self.patternlist)
        self.bToolsView.clicked.connect(self.tools)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.popupmenu = QMenu()
        self.diagview = None
        lmodel = self.FG.model
        self.treeview.setModel(lmodel)
        self.treeview.setItemDelegate(Q7TreeItemDelegate(self.treeview, lmodel))
        self.treeview.setControlWindow(self, self.FG.index)
        if (self._control.transientRecurse or OCTXT.RecursiveTreeDisplay):
            self.expandMinMax()
        if (self._control.transientVTK): self.vtkview()
        self._control.transientRecurse = False
        self._control.transientVTK = False
        self.clearchecks()
        #
        self.bCheckList.setDisabled(True)
        if (not OCTXT._HasProPackage):
            self.bToolsView.setDisabled(True)
        self.bCheckView.setDisabled(True)
        self.bPatternDB.setDisabled(True)
        self.bAddLink.clicked.connect(self.linkadd)
        self.bSelectLinkSrc.clicked.connect(self.linkselectsrc)
        self.bSelectLinkDst.clicked.connect(self.linkselectdst)
        self.bAddLink.setDisabled(True)
        self.lineEdit.returnPressed.connect(self.jumpToNode)
        # QObject.connect(self.lineEdit,
        #                SIGNAL("returnPressed()"),
        #                self.jumpToNode)
        tvh = self.treeview.header()
        tvh.setContextMenuPolicy(Qt.CustomContextMenu)
        tvh.customContextMenuRequested.connect(self.headerMenu)
        self._hmenu = QMenu()
        self._hmenu._idx = {}
        self._tlist = (('SIDS type', NMT.COLUMN_SIDS),
                       ('Link flag', NMT.COLUMN_FLAG_LINK),
                       ('Mark flag', NMT.COLUMN_FLAG_SELECT),
                       ('Check flag', NMT.COLUMN_FLAG_CHECK),
                       ('User flag', NMT.COLUMN_FLAG_USER),
                       ('Shape', NMT.COLUMN_SHAPE),
                       ('Data type', NMT.COLUMN_DATATYPE))
        for (tag, idx) in self._tlist:
            a = QAction(tag, self._hmenu, checkable=True)
            self._hmenu._idx[idx] = a
            if (self._column[idx]):
                a.setChecked(True)
            else:
                a.setChecked(False)
            self._hmenu.addAction(a)
            self.treeview.setColumnHidden(idx, not self._column[idx])
        self._recursiveAddNewNode = False
        self.updateTreeStatus()

    def headerMenu(self, pos):
        for (tag, idx) in self._tlist:
            self._hmenu._idx[idx].setChecked(self._column[idx])
        self._hmenu.exec_(self.treeview.mapToGlobal(pos))
        for (tag, idx) in self._tlist:
            if (self._hmenu._idx[idx].isChecked()):
                self._column[idx] = True
            else:
                self._column[idx] = False
            self.treeview.setColumnHidden(idx, not self._column[idx])

    def model(self):
        return self.FG.model

    def modelIndex(self, idx):
        if not idx.isValid():
            return -1
        midx = idx
        if idx.model() != self.treeview.M():
            midx = self.treeview.model().mapToSource(idx)
        return midx

    def modelData(self, idx):
        if not idx.isValid():
            return None
        return self.modelIndex(idx).internalPointer()

    def savetree(self):
        if (not (self.FG.isSaveable() and self.FG.isModified())):
            return
        self._control.savedirect(self.FG)
        self.updateTreeStatus()

    def tools(self):
        from CGNS.NAV.wtools import Q7ToolsView
        if (self._control._toolswindow is None):
            self._control._toolswindow = Q7ToolsView(self._control,
                                                     self.FG,
                                                     self)
            self._control._toolswindow.show()
        else:
            self._control._toolswindow.raise_()

    def savetreeas(self):
        self._control.save(self.FG)
        self.updateTreeStatus()

    def infoTreeView(self):
        self._control.helpWindow('Tree')

    def screenshot(self):
        self.treeview.model().sort(0)
        sshot = QScreen.grabWindow(self.treeview.winId())
        sshot.save('/tmp/foo.png', 'png')

    def expandMinMax(self):
        if (self._depthExpanded == self.FG.depth - 2):
            self._depthExpanded = -1
            self.treeview.collapseAll()
        else:
            self._depthExpanded = self.FG.depth - 2
            self.treeview.expandAll()
        self.resizeAll()

    def resetOptions(self):
        if (OCTXT.AutoExpand):
            self.treeview.setAutoExpandDelay(1000)
        else:
            self.treeview.setAutoExpandDelay(-1)

    def expandLevel(self):
        if (self._depthExpanded < self.FG.depth - 2):
            self._depthExpanded += 1
        self.treeview.expandToDepth(self._depthExpanded)
        self.resizeAll()

    def collapseLevel(self):
        if (self._depthExpanded != -1): self._depthExpanded -= 1
        if (self._depthExpanded == -1):
            self.treeview.collapseAll()
        else:
            self.treeview.expandToDepth(self._depthExpanded)
        self.resizeAll()

    def updateStatus(self, node):
        if (not self.lineEditLock.isChecked()):
            self.lineEdit.clear()
            self.lineEdit.insert(node.sidsPath())

    def jumpToNode(self):
        path = self.lineEdit.text()
        self.treeview.selectByPath(path)

    def popform(self):
        self.formview()

    def openLkTree(self):
        self.busyCursor()
        filename = self.getLastEntered().sidsLinkFilename()
        if (filename is not None):
            self._control.loadfile(filename)
        self.readyCursor()

    def openSubTree(self):
        self.busyCursor()
        node = self.getLastEntered().sidsPath()
        child = Q7Tree(self._control, node, self.FG)
        self.readyCursor()
        child.show()

    def pop0(self):
        pass

    def newnodebrother(self):
        if (self.getLastEntered() is not None):
            self.model().newNodeBrother(self.getLastEntered())

    def newnodechild(self):
        if (self.getLastEntered() is not None):
            self.model().newNodeChild(self.getLastEntered())

    def marknode(self):
        if (self.getLastEntered() is not None):
            self.treeview.markNode(self.getLastEntered())

    def mcopy(self):
        if (self.getLastEntered() is not None):
            self.model().copyNode(self.getLastEntered())
            self.clearOtherSelections()

    def mcutselected(self):
        self.model().cutAllSelectedNodes()
        self.clearLastEntered()
        self.clearOtherSelections()

    def mcut(self):
        if (self.getLastEntered() is not None):
            self.model().cutNode(self.getLastEntered())
            self.clearLastEntered()
            self.clearOtherSelections()

    def mdel(self):
        if (self.getLastEntered() is not None):
            self.model().deleteNode(self.getLastEntered())
            self.clearLastEntered()
            self.clearOtherSelections()

    def mpasteasbrotherselected(self):
        self.model().pasteAsBrotherAllSelectedNodes()

    def mpasteasbrother(self):
        if (self.getLastEntered() is not None):
            self.model().pasteAsBrother(self.getLastEntered())

    def mpasteaschildselected(self):
        self.model().pasteAsChildAllSelectedNodes()

    def mpasteaschild(self):
        if (self.getLastEntered() is not None):
            self.model().pasteAsChild(self.getLastEntered())

    def updateMenu(self, nodeidxs):
        nodeidx = self.modelIndex(nodeidxs)
        if (not nodeidx.isValid): return False
        if (nodeidx.internalPointer() is None): return False
        if (nodeidx.internalPointer().sidsPath() == '/CGNSTree'): return False
        self.setLastEntered(nodeidxs)
        if (nodeidx != -1):
            node = nodeidx.internalPointer()
            lknode = not node.sidsIsLink()
            lznode = node.hasLazyLoad()
            actlist = (
                ("%s goodies" % node.sidsType(),),
                None,
                ("Expand sub-tree from this node", self.expand_sb, 'Ctrl++', False),
                ("Collapses sub-tree from this node", self.collapse_sb, 'Ctrl+-', False),                
                None,
                ['Mark nodes...',[
                ("Mark/unmark node", self.marknode, 'Space', False),
                 None,
                ("Mark all nodes same SIDS type", self.marknode_t, 'Ctrl+1', False),
                ("Mark all nodes same name", self.marknode_n, 'Ctrl+2', False),
                ("Mark all nodes same value", self.marknode_v, 'Ctrl+3', False),
                 None,
                ("Mark parent path", self.marknode_p, 'Ctrl+4', False)]],
                ("Add new child node", self.newnodechild, 'Ctrl+A', False),
                ("Add new brother node", self.newnodebrother, 'Ctrl+Z', False),
                #            None,
                #            ("Open form",self.popform,'Ctrl+F',False),
                #            ("Open view",self.openSubTree,'Ctrl+W',False),
                #            ("Open view on linked-to file",self.openLkTree,'Ctrl+O',lknode),
                None,
                ("Load node data in memory", self.dataLoad, 'Ctrl+L', not lznode),
                ("Release memory node data", self.dataRelease, 'Ctrl+R', lznode),
                None,
                ("Copy current", self.mcopy, 'Ctrl+C', False),
                ("Cut current", self.mcut, 'Ctrl+X', False),
                ("Paste as brother", self.mpasteasbrother, 'Ctrl+V', False),
                ("Paste as child", self.mpasteaschild, 'Ctrl+Y', False),
                ("Delete current", self.mdel, QKeySequence(Qt.Key_Delete), False),
                None,
                ['On selected nodes...',[
                ("Expand sub-tree from all selected nodes", self.sexpand_sb, 'Ctrl+Shift++', False),
                ("Collapses sub-tree from all selected nodes", self.scollapse_sb, 'Ctrl+Shift+-', False),
                None,
                ("Cut all selected", self.mcutselected, 'Ctrl+Shift+X', False),
                ("Paste as brother for each selected",
                 self.mpasteasbrotherselected, 'Ctrl+Shift+V', False),
                ("Paste as child for each selected",
                 self.mpasteaschildselected, 'Ctrl+Shift+Y', False),
                 None,
                ("Load nodes data in memory for each selected",
                 self.dataLoadSelected, 'Ctrl+Shift+L', False),
                ("Release memory node data for each selected",
                 self.dataReleaseSelected, 'Ctrl+Shift+R', False)]],
            )
            self.popupmenu.clear()
            self.popupmenu.setTitle('Node menu')
            for aparam in actlist:
                if (aparam is None):
                    self.popupmenu.addSeparator()
                elif (len(aparam) == 1):
                    stp = node.sidsType()
                    tag = '_GM_{}'.format(stp)
                    subm = self.popupmenu.addMenu('{}...'.format(stp))
                    a = QAction("About %s" % node.sidsType(), self,
                                triggered=self.aboutSIDS)
                    subm.addAction(a)
                    patmenu = subm.addMenu('Insert pattern')
                    self.patternMenu(patmenu, node.sidsNode())
                    subm.addSeparator()
                    if (hasattr(self, tag)):
                        getattr(self, tag)(subm, node)
                else:
                    if isinstance(aparam,list):
                        subm = self.popupmenu.addMenu(aparam[0])
                        for aaparam in aparam[1]:
                            if (aaparam is None):
                                 subm.addSeparator()
                            else:
                                a = QAction(aaparam[0], self, triggered=aaparam[1])
                                if (aaparam[2] is not None): a.setShortcut(aaparam[2])
                                subm.addAction(a)
                                a.setDisabled(aaparam[3])
                    else:
                        a = QAction(aparam[0], self, triggered=aparam[1])
                        if (aparam[2] is not None): a.setShortcut(aparam[2])
                        self.popupmenu.addAction(a)
                        a.setDisabled(aparam[3])
            return True

    def _runAndSelect(self, qname, value):
        q = Q7Query.getQuery(qname)
        sl = q.run(self.FG.tree, self.FG.links, list(self.FG.lazy), False, value)
        self.model().markExtendToList(sl)
        self.model().updateSelected()
        self.treeview.refreshView()

    def patternMenu(self, menu, node):
        a = QAction("Recursive sub-pattern add", self, checkable=True)
        menu.addAction(a)
        a.setChecked(self._recursiveAddNewNode)
        menu.addSeparator()
        for t in [n[0] for n in CGU.getAuthChildren(node)]:
            def genCopyPattern(arg):
                def copyPattern():
                    pattern = CGS.profile.get(arg, None)
                    if pattern is None:
                       str_msg ="Pattern for {} was not found in current profile".format(arg)
                       MSG.wError(self, 404, 'Cannot create recursive sub-pattern', str_msg)
                       return
                    self.model().copyNodeRaw(pattern[0])
                    if (self.getLastEntered() is not None):
                        self.model().pasteAsChild(self.getLastEntered())
                return copyPattern
            a = QAction("{}".format(t), self, triggered=genCopyPattern(t))
            menu.addAction(a)
        
    def _gm_family_1(self, node):
        self._runAndSelect('013. FamilyName reference', "'%s'" % node.sidsName())

    def _gm_family_2(self, node):
        self._runAndSelect('003. Node type', "'Family_t'")

    def _GM_Family_t(self, m, node):
        a = QAction('Select references to myself', self)
        a.triggered.connect(functools.partial(self._gm_family_1, node))
        m.addAction(a)
        a = QAction('Select all families', self)
        a.triggered.connect(functools.partial(self._gm_family_2, node))
        m.addAction(a)
        m.addSeparator()
        return True

    def _GM_IndexRange_t(self, m, node):
        if (node.sidsName() != CGK.ElementRange_s):
            v = 0
            a = QAction('Range size: %d' % (v), self)
            m.addAction(a)
        else:
            v = node.sidsValue()[1] - node.sidsValue()[0]
            etp = CGU.getEnumAsString(node.sidsParent())
            a = QAction('Number of elements of type [%s]: %d' % (etp, v), self)
            m.addAction(a)
        return True

    def _GM_Elements_t(self, m, node):
        etp = CGU.getEnumAsString(node.sidsNode())
        npe = CGK.ElementTypeNPE[etp]
        a = QAction('Element type [%s] npe [%d]' % (etp, npe), self)
        m.addAction(a)
        return True

    def marknode_t(self):
        node = self.getLastEntered()
        self._runAndSelect('003. Node type', "'%s'" % node.sidsType())

    def marknode_n(self):
        node = self.getLastEntered()
        self._runAndSelect('001. Node name', "'%s'" % node.sidsName())

    def marknode_v(self):
        node = self.getLastEntered()
        value = node.sidsValue()
        self._runAndSelect('005. Node value', value)

    def marknode_p(self):
        node = self.getLastEntered()
        node.switchMarked()
        path = node.sidsPath()
        while path is not None:
            path = CGU.getPathAncestor(path)
            if (path not in ['/', None]):
                node = self.model().nodeFromPath('/CGNSTree' + path)
                node.switchMarked()
        self.model().updateSelected()

    def setLastEntered(self, nix=None):
        if ((nix is None) or (not nix.isValid())):
            nix = self.treeview.modelCurrentIndex()
        self._lastEntered = None
        if (nix.isValid()):
            self.treeview.exclusiveSelectRow(nix, False)
            self._lastEntered = self.modelData(nix)

    def getLastEntered(self):
        return self._lastEntered

    def clearLastEntered(self):
        self._lastEntered = None
        self.treeview.selectionModel().clearSelection()
        return None

    def clearOtherSelections(self):
        if (self._control._patternwindow is not None):
            self._control._patternwindow.clearSelection()

    def clickedPressedNode(self, index):
        self.clickedNode(index)

    def clickedNode(self, index):
        self.treeview.exclusiveSelectRow(index)
        if (self.treeview.lastButton == Qt.RightButton):
            if (self.updateMenu(index)):
                self.popupmenu.popup(self.treeview.lastPos)

    def expandNode(self, *args):
        self.resizeAll()

    def collapseNode(self, *args):
        pass

    def expand_sb(self):
        self.treeview.expand_sb()
    
    def collapse_sb(self):
        self.treeview.collapse_sb()
    
    def sexpand_sb(self):
        self.treeview.sexpand_sb()
    
    def scollapse_sb(self):
        self.treeview.scollapse_sb()
    
    def resizeAll(self):
        for n in range(NMT.COLUMN_LAST + 1):
            self.treeview.resizeColumnToContents(n)

    def show(self):
        super(Q7Tree, self).show()

    def linkselectsrc(self):
        if (self.bSelectLinkSrc.isChecked()):
            if (self.getLastEntered() is None): return
            self.bAddLink.setDisabled(False)
            node = self.getLastEntered()
            self.selectForLinkSrc = (node, node.sidsPath())
        else:
            self.bAddLink.setDisabled(True)
            self.selectForLinkSrc = None

    def linkselectdst(self):
        if (self.getLastEntered() is None): return
        node = self.getLastEntered()
        if (node is None):  return
        if (node.sidsIsLink()): return
        if (node.sidsType() == CGK.CGNSTree_ts): return
        if (self._control.selectForLinkDst is not None):
            bt = self._control.selectForLinkDst[-1].bSelectLinkDst
            bt.setChecked(Qt.Unchecked)
            if (self._control.selectForLinkDst[-1] == self):
                self._control.selectForLinkDst = None
                return
        self._control.selectForLinkDst = (node, node.sidsPath(),
                                          self.FG.filedir,
                                          self.FG.filename,
                                          self)
        self.bSelectLinkDst.setChecked(Qt.Checked)
        if (self._linkwindow is not None):
            n = node.sidsPath()
            d = self.FG.filedir
            f = self.FG.filename
            self._linkwindow.updateSelected(d, f, n)

    def linkadd(self):
        if (self._control.selectForLinkDst is None): return
        dst = self._control.selectForLinkDst
        str_dst = "%s:%s" % (dst[3], dst[1])
        tpath = 'relative'
        newname = CGU.getPathLeaf(dst[1])
        if (CGU.checkDuplicatedName(self.selectForLinkSrc[0].sidsNode(),
                                    newname, dienow=False)):
            str_cnm = "New child node name is <b>%s</b>" % newname
        else:
            count = 0
            while (not CGU.checkDuplicatedName(self.selectForLinkSrc[0].sidsNode(),
                                               newname, dienow=False)):
                count += 1
                newname = '{%s#%.3d}' % (dst[0].sidsType(), count)
            str_cnm = """As a child with this name already exists, the name <b>%s</b> is used (generated name)""" % \
                      newname
        str_src = "%s:%s/%s" % (self.FG.filename,
                                self.selectForLinkSrc[1], newname)
        str_msg = "you want to create a link from <b>%s</b> to <b>%s</b><br>%s<br>Your current user options do force " \
                  "the link to use <b>%s</b> destination file path.""" % (
        str_src, str_dst, str_cnm, tpath)
        reply = MSG.wQuestion(self, 231, 'Create link as a new node', str_msg)

    def linklist(self):
        if (self._linkwindow is None):
            self._linkwindow = Q7LinkList(self._control, self.FG.index, self)
            self._linkwindow.show()
        else:
            self._linkwindow.raise_()

    def patternlist(self):
        if (self._control._patternwindow is None):
            self._control._patternwindow = Q7PatternList(self._control, self.FG)
            self._control._patternwindow.show()
        self._control._patternwindow.raise_()

    def check(self):
        self.busyCursor()
        if (self.diagview is not None):
            self.diagview.close()
            self.diagview = None
        self.lastdiag = self.model().checkSelected()
        self.readyCursor()
        self.treeview.refreshView()
        self.bCheckList.setDisabled(False)

    def checklist(self):
        if (self.lastdiag is None): return
        self.diagview = Q7CheckList(self, self.lastdiag, self.FG.index)
        self.diagview.show()

    def clearchecks(self):
        self.model().checkClear()
        self.treeview.refreshView()
        self.lastdiag = None
        self.bCheckList.setDisabled(True)

    def selectionlist(self):
        if (self._selectwindow is not None):
            self._selectwindow.close()
            self._selectwindow = None
        self._selectwindow = Q7SelectionList(self, self.model(), self.FG.index)
        self._selectwindow.show()
        self._selectwindow.raise_()

    def previousmark(self):
        self.treeview.changeSelectedMark(-1)

    def nextmark(self):
        self.treeview.changeSelectedMark(+1)

    def markall(self):
        self.model().markAll()
        self.model().updateSelected()
        self.treeview.refreshView()

    def unmarkall(self):
        self.model().unmarkAll()
        self.model().updateSelected()
        self.treeview.refreshView()

    def swapmarks(self):
        self.model().swapMarks()
        self.model().updateSelected()
        self.treeview.refreshView()

    def formview(self):
        ix = self.treeview.modelCurrentIndex()
        node = self.modelData(ix)
        if (node is None):
            MSG.wInfo(self, 254, "Form view:",
                      """You have to select a node to open its form view""",
                      again=False)
            return
        if (node.sidsType() == CGK.CGNSTree_ts): return
        form = Q7Form(self._control, node, self.FG.index)
        form.show()

    def vtkview(self):
        if (not HAS_VTK): return
        from CGNS.NAV.wvtk import Q7VTK
        if (self._vtkwindow is None):
            self.busyCursor()
            ix = self.treeview.modelCurrentIndex()
            zlist = self.model().getSelectedZones()
            node = self.modelData(ix)
            self._vtkwindow = Q7VTK(self._control, self, node, self.FG.index,
                                    self.model(), zlist)
            if (self._vtkwindow._vtkstatus):
                self._vtkwindow.show()
            else:
                self._vtkwindow.close()
                self._vtkwindow = None
            self.readyCursor()
        else:
            self._vtkwindow.raise_()

    def plotview(self):
        return

    def queryview(self):
        if (self._querywindow is None):
            self._querywindow = Q7Query(self._control, self.FG.index, self)
            self._querywindow.show()
        else:
            self._querywindow.raise_()

    def aboutSIDS(self):
        path = self.getLastEntered().sidsPath()

    def dataLoadSelected(self):
        self.model().dataLoadSelected()

    def dataReleaseSelected(self):
        self.model().dataReleaseSelected()

    def dataLoad(self):
        node = self.getLastEntered()
        self.model().dataLoadSelected(single=node.sidsPath())

    def dataRelease(self):
        node = self.getLastEntered()
        self.model().dataReleaseSelected(single=node.sidsPath())

    def forceapply(self):
        pass

    def updateTreeStatus(self):
        if ((Q7FingerPrint.STATUS_MODIFIED in self.FG._status)
            and (Q7FingerPrint.STATUS_SAVEABLE in self.FG._status)):
            self.bSave.setEnabled(True)
        else:
            self.bSave.setEnabled(False)

    def doRelease(self):
        # break cyclic refs to allow garbage
        self.treeview.itemDelegate().doRelease()
        self.treeview.setItemDelegate(None)
        self.treeview.doRelease()
        self.treeview = None

# -----------------------------------------------------------------



#  -------------------------------------------------------------------------
#  pyCGNS.VAL - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
import CGNS.VAL.grammars.CGNS_VAL_USER_DEFAULT as CGV
import CGNS.VAL.parse.messages as CGM
import CGNS.PAT.cgnserrors     as CGE
import CGNS.VAL.parse.findgrammar

import sys
import os


def listuserkeys(trace):
    s = ''
    if not trace:
        print('### Use -v (verbose) option to check if the found paths')
        print('### are into your PYTHONPATH. If not, grammar would not be')
        print("### used by CGNS.VAL even if detected...")
    dk = CGNS.VAL.parse.findgrammar.findAllUserGrammars(trace)
    for key in dk:
        s += '%-16s: %s\n' % (key, dk[key])
    return s


def getParser(trace, user):
    if user is not None:
        mod = CGNS.VAL.parse.findgrammar.importUserGrammars(user, verbose=trace)
        if mod is None:
            if trace:
                print('### Using grammar [DEFAULT]')
            parser = CGV.CGNS_VAL_USER_Checks(None)
        else:
            if trace:
                print('### Using grammar [%s]' % user)
            if trace:
                print('### Found in [%s]' % os.path.dirname(mod.__file__))
            parser = mod.CGNS_VAL_USER_Checks(None)
    else:
        parser = CGV.CGNS_VAL_USER_Checks(None)
    return parser


def listdiags(trace, userlist):
    ld = []
    for user in userlist:
        parser = getParser(trace, user)
        mlist = parser.listDiagnostics()
        ld += list(mlist)
    ld.sort()
    s = ''
    for d in ld:
        s += "%s\n" % (str(mlist[d]))
    return s


def run(T, trace, userlist, stop=False, warnings=[], failures=[]):
    diag = CGM.DiagnosticLog()
    for user in userlist:
        parser = getParser(trace, user)
        # parser lookup and init are required prior to initialize message tables
        for w in warnings:
            diag.forceAsWarning(w)
        for w in failures:
            diag.forceAsFailure(w)
        try:
            parser.checkTree(T, trace, stop=stop)
        except (CGE.cgnsException,) as v:
            pass
        diag.merge(parser.log)
    return diag


def compliant(T, trace=False, userlist=['DEFAULT'], paths=[''], stop=False,
              warnings=[], failures=[]):
    ipath = '%s/lib/python%s.%s/site-packages/CGNS/VAL/grammars' % \
            (sys.prefix, sys.version_info[0], sys.version_info[1])
    sys.path.append(ipath)
    for pp in paths: sys.path.append(pp)
    diag = run(T, trace, userlist, stop=stop, warnings=warnings, failures=failures)
    ok = [True, []]
    for p in diag:
        for (s, sp) in diag.diagnosticsByPath(p):
            if (diag.status(s) == CGM.CHECK_FAIL):
                ok[0] = False
                ok[1].append((p, diag.message(s)))
    return ok


def showDiag(diag, idlist, bypath=True):
    ok = True
    if bypath:
        for p in diag:
            if not diag.hasOnlyKey(p, idlist):
                print('\n%s\n%s' % ('-' * 75, p))
                for (s, sp) in diag.diagnosticsByPath(p):
                    if ((diag.status(s) != CGM.CHECK_GOOD) and
                            (diag.key(s) not in idlist)):
                        print(diag.message(s))
                        if diag.status(s) == CGM.CHECK_FAIL:
                            ok = False
        print('\n%s\n' % ('-' * 75))
    else:
        for m in diag.allMessageKeys():
            if (m not in idlist):
                first = True
                ctxt = diag.noContextMessage(m)
                if (ctxt is not None):
                    print('\n%s\n[%s] %s' % ('-' * 75, m, ctxt))
                else:
                    print('\n%s\n[%s]' % ('-' * 75, m))
                for (d, dp) in diag.diagnosticsByMessage(m):
                    if diag.status(d) != CGM.CHECK_GOOD:
                        if ctxt is None:
                            if not first:
                                skip = '\n'
                            else:
                                skip = ''
                            print('%s  %s\n  > %s' % (skip, dp, d.message))
                        else:
                            print('  %s' % (dp))
                        first = False
                        if diag.status(d) == CGM.CHECK_FAIL:
                            ok = False
        print('\n%s' % ('-' * 75))
    if ok:
        print('### CGNS/Python tree Compliant')
    else:
        print('### CGNS/Python tree *NOT* Compliant')



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

from CGNS.NAV.moption import Q7OptionContext as OCTXT

import sys
import string
import re

from qtpy.QtWidgets import QTextEdit, QDialog
from qtpy.QtGui import QTextCursor, QSyntaxHighlighter, QTextCharFormat, QFont, QColor


from CGNS.NAV.Q7MessageWindow import Ui_Q7MessageWindow
from CGNS.NAV.Q7LogWindow import Ui_Q7LogWindow

(INFO, QUESTION, ERROR, WARNING) = (0, 1, 2, 3)

# globals
OK = True
CANCEL = False
ANSWER = True


# -----------------------------------------------------------------
class Q7DocEditor(QTextEdit):
    def __init__(self, parent=None):
        QTextEdit.__init__(self, parent)

    def initText(self, text):
        self.clear()
        self.insertPlainText(text)
        self.moveCursor(QTextCursor.Start, QTextCursor.MoveAnchor)
        self.ensureCursorVisible()


# -----------------------------------------------------------------
class Q7PythonEditor(QTextEdit):
    def __init__(self, parent=None):
        QTextEdit.__init__(self, parent)
        self.highlighter = Q7PythonEditorHighlighter(self.document())
        fsz = 12
        ffm = 'Courier'
        self.setStyleSheet("font: %dpt \"%s\";" % (fsz, ffm))

    def initText(self, text):
        self.clear()
        self.insertPlainText(text)
        self.moveCursor(QTextCursor.Start, QTextCursor.MoveAnchor)
        self.ensureCursorVisible()


# -----------------------------------------------------------------
class Q7PythonEditorHighlighter(QSyntaxHighlighter):
    def __init__(self, *args):
        keywords = r"\bimport\b|\bas\b|\bfor\b|\bdef\b|\bpass\b"
        keywords += r"|\bclass\b|\bfrom\b|\bif\b|\bthen\b|\belse\b"
        keywords += r"|\belif\b|\btry\b|\bexcept\b|\bfinally\b|\braise\b"
        keywords += r"|\bprint\b|\bin\b|\bnot\b|\band\b|\bor\b|\bcontinue\b"
        keywords += r"|\bwhile\b|\breturn\b|\blambda\b|\bwith\b|\del\b"
        keywords += r"|\bglobal\b|\byield\b|\bexec\b|\bassert\b|\break\b"
        constants = r"\bNone\b|\bTrue\b|\bFalse\b|\bself\b"
        autovars = r"\bNODE\b|\bNAME\b|\bVALUE\b|\bSIDSTYPE\b|\bCHILDREN\b"
        autovars += r"|\bTREE\b|\bPATH\b|\bRESULT\b|\bARGS\b|\bPARENT\b"
        autovars += r"|\bLINKS\b|\bSKIPS\b"
        numbers = r'[-+]?\d+' + '|[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?'
        comment = r'\#.*$'
        textstring = r'"[^"]*?"|\'[^\']*?\''
        self.f_keywords = QTextCharFormat()
        self.f_keywords.setFontWeight(QFont.Bold)
        self.f_keywords.setForeground(QColor('Blue'))
        self.f_comment = QTextCharFormat()
        self.f_comment.setFontWeight(QFont.Light)
        self.f_comment.setForeground(QColor('Red'))
        self.f_constants = QTextCharFormat()
        self.f_constants.setFontWeight(QFont.Light)
        self.f_constants.setForeground(QColor('Navy'))
        self.f_autovars = QTextCharFormat()
        self.f_autovars.setFontWeight(QFont.Bold)
        self.f_autovars.setForeground(QColor('Green'))
        self.f_textstring = QTextCharFormat()
        self.f_textstring.setFontWeight(QFont.Bold)
        self.f_textstring.setForeground(QColor('Coral'))
        self.f_numbers = QTextCharFormat()
        self.f_numbers.setFontWeight(QFont.Light)
        self.f_numbers.setForeground(QColor('Gray'))
        self.r_keywords = re.compile(keywords)
        self.r_numbers = re.compile(numbers)
        self.r_comment = re.compile(comment)
        self.r_constants = re.compile(constants)
        self.r_autovars = re.compile(autovars)
        self.r_textstring = re.compile(textstring)
        self.in_comment = False
        self.in_string = False
        self.r_syntax = []
        self.r_syntax.append((self.r_keywords, self.f_keywords))
        self.r_syntax.append((self.r_numbers, self.f_numbers))
        self.r_syntax.append((self.r_autovars, self.f_autovars))
        self.r_syntax.append((self.r_constants, self.f_constants))
        self.r_syntax.append((self.r_comment, self.f_comment))
        self.r_syntax.append((self.r_textstring, self.f_textstring))
        QSyntaxHighlighter.__init__(self, *args)

    def highlightBlock(self, textblock):
        text = str(textblock)
        fdef = self.format(0)
        fmax = len(text)
        for (rex, fmt) in self.r_syntax:
            mit = rex.finditer(text)
            for m in mit:
                self.setFormat(m.span()[0], m.span()[1], fmt)
                self.setFormat(m.span()[1], fmax, fdef)


# -----------------------------------------------------------------
class Q7Log(QDialog, Ui_Q7LogWindow):
    def __init__(self):
        QDialog.__init__(self, None)
        self.setupUi(self)
        self.bClose.clicked.connect(self.leave)
        self.bClear.clicked.connect(self.clear)
        self.setWindowTitle("%s: Log" % OCTXT._ToolName)
        self.eLog.setReadOnly(True)
        self.eLog.setAcceptRichText(False)
        self.eLog.setStyleSheet("font: 12pt \"Courier\";")

    def write(self, msg):
        self.eLog.insertPlainText(msg)

    def clear(self):
        self.eLog.clear()

    def leave(self, *arg):
        self.hide()


# -----------------------------------------------------------------
class Q7MessageBox(QDialog, Ui_Q7MessageWindow):
    def __init__(self, control, code):
        QDialog.__init__(self, None)
        self.setupUi(self)
        self.bOK.clicked.connect(self.runOK)
        self.bCANCEL.clicked.connect(self.runCANCEL)
        self.bInfo.setDisabled(True)
        self._text = ''
        self._control = control
        self._code = 0

    def setMode(self, cancel=False, again=False):
        if (not again): self.cNotAgain.hide()
        if (not cancel): self.bCANCEL.hide()

    def setLayout(self, text, btype=INFO,
                  cancel=False, again=False, buttons=('OK', 'Cancel')):
        self.bOK.setText(buttons[0])
        if len(buttons) > 1:
            self.bCANCEL.setText(buttons[1])
        self.eMessage.setText(str(text))
        self._text = text
        self.eMessage.setReadOnly(True)
        self.setMode(cancel, again)

    def runOK(self, *arg):
        self.addToSkip()
        self.close()
        self.setResult(QDialog.Accepted)

    def runCANCEL(self, *arg):
        self.addToSkip()
        self.close()
        self.setResult(QDialog.Rejected)

    def addToSkip(self):
        if (self.cNotAgain.isChecked()):
            if (self._code and self._code not in OCTXT.IgnoredMessages):
                OCTXT.IgnoredMessages += [self._code]

    def showAndWait(self):
        ret = self.exec_()
        if ret == QDialog.Accepted:
            return True
        else:
            return False


def wError(control, code, info, error):
    txt = """<img source=":/images/icons/user-G.png">  <big>ERROR #%d</big><hr>
         %s<br>%s""" % (code, error, info)
    if (code in OCTXT.IgnoredMessages): return True
    msg = Q7MessageBox(control, code)
    msg.setWindowTitle("%s: Error" % OCTXT._ToolName)
    msg.setLayout(txt, btype=ERROR, cancel=False, again=True, buttons=('Close',))
    return msg.showAndWait()


def wQuestion(control, code, title, question, again=True, buttons=('OK', 'Cancel')):
    txt = """<img source=":/images/icons/user-M.png">
         <b> <big>%s</big></b><hr>%s""" % (title, question)
    if (code in OCTXT.IgnoredMessages): return True
    msg = Q7MessageBox(control, code)
    msg.setWindowTitle("%s: Question" % OCTXT._ToolName)
    msg.setLayout(txt, btype=QUESTION, cancel=True, again=again, buttons=buttons)
    return msg.showAndWait()


def wInfo(control, code, title, info, again=True, buttons=('Close',)):
    txt = """<img source=":/images/icons/user-S.png">
         <b> <big>%s</big></b><hr>%s""" % (title, info)
    if (code in OCTXT.IgnoredMessages): return True
    msg = Q7MessageBox(control, code)
    msg.setWindowTitle("%s: Info" % OCTXT._ToolName)
    msg.setLayout(txt, btype=INFO, cancel=False, again=again, buttons=buttons)
    return msg.showAndWait()

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
from builtins import (str, bytes, range, dict, type)

from CGNS.NAV.moption import Q7OptionContext as OCTXT

import os
import os.path
import time
import stat
import sys

import CGNS.MAP
import CGNS.PAT.cgnsutils as CGU

from qtpy.QtCore import QCoreApplication
from qtpy.QtCore import Qt, QThread, QMutex, Signal
from qtpy.QtWidgets import QWidget
from qtpy.QtGui import (QIcon, QPixmap, QFont, QCursor)

from CGNS.NAV.wstylesheets import Q7TREEVIEWSTYLESHEET, Q7TABLEVIEWSTYLESHEET
from CGNS.NAV.wstylesheets import Q7CONTROLVIEWSTYLESHEET
from CGNS.NAV.wfile import checkFilePermission

import CGNS.NAV.wmessages as MSG

try:
    from pwd import getpwuid
except ImportError:
    def getpwuid(a):
        return (0,)
try:
    from grp import getgrgid
except ImportError:
    def getgrgid(a):
        return (0,)


# -----------------------------------------------------------------
class Q7CHLoneProxy(object):
    def __init__(self, hasthreads=False):
        self._data = None
        self._hasthreads = hasthreads
        self._thread = None

    def kill(self):
        self._data = None
        if self._thread is not None:
            self._thread.quit()

    def load(self, control, selectedfile):
        self._control = control
        self._thread = Q7CHLoneThread(control, selectedfile)
        self._thread.datacompleted.connect(self.proxyCompleted,Qt.QueuedConnection)
        self._data = None
        if self._hasthreads:
            self._thread.start()
        else:
            self._thread.run()

    @property
    def data(self):
        return self._data

    # not useful anymore ?
    #@Slot(tuple, name='proxyCompleted')
    def proxyCompleted(self, data):
        if data[0] is None:
            self._data = data
        else:
            self._data = data[0]
        self._control.signals.fgprint = self._data
        self._control.signals.loadCompleted.emit()


# -----------------------------------------------------------------
class Q7CHLoneThread(QThread):
    datacompleted = Signal(tuple, name='datacompleted')

    def __init__(self, control, selectedfile):
        super(Q7CHLoneThread, self).__init__(None)
        self._control = control
        self._selectedfile = selectedfile
        self._data = None

    def run(self):
        control = self._control
        selectedfile = str(self._selectedfile)
        Q7FingerPrint.Lock()
        kw = {}
        (filedir, filename) = (os.path.abspath(os.path.dirname(selectedfile)),
                               os.path.basename(selectedfile))
        if "%s/%s" % (filedir, filename) in Q7FingerPrint.getExpandedFilenameList():
            Q7FingerPrint.Unlock()
            txt = """The current file is already open:"""
            self._data = (None, (0, "%s/%s" % (filedir, filename), txt))
            self.datacompleted.emit(self._data)
            control.readyCursor()
            return
        slp = OCTXT.LinkSearchPathList
        slp += [filedir]
        slp = [str(lp) for lp in slp]
        loadfilename = selectedfile
        flags = CGNS.MAP.S2P_DEFAULTS
        maxdataload = -1
        if OCTXT.CHLoneTrace:
            flags |= CGNS.MAP.S2P_TRACE
        if OCTXT.DoNotLoadLargeArrays:
            flags |= CGNS.MAP.S2P_NODATA
            maxdataload = OCTXT.MaxLoadDataSize
        if OCTXT.FollowLinksAtLoad:
            flags |= CGNS.MAP.S2P_FOLLOWLINKS
        try:
            if not CGNS.MAP.probe(loadfilename):
                if ((os.path.splitext(filename)[1] in OCTXT.CGNSFileExtension) and
                        OCTXT._ConvertADFFiles):
                    loadfilename = Q7FingerPrint.fileconversion(filedir,
                                                                filename, control)
                    kw['converted'] = True
                    kw['convertedAs'] = loadfilename
            if maxdataload:
                (tree, links, paths) = CGNS.MAP.load(loadfilename, flags=flags,
                                                     lksearch=slp,
                                                     maxdata=maxdataload)
            else:
                (tree, links, paths) = CGNS.MAP.load(loadfilename, flags=flags,
                                                     lksearch=slp)
        except CGNS.MAP.error as chlex:
            Q7FingerPrint.Unlock()
            txt = """Load aborted by CGNS.MAP (file:%s)"""%loadfilename
            code = chlex.args[0][0]
            msg = chlex.args[0][1]
            self._data = (None, (code, msg, txt))
            self.datacompleted.emit(self._data)
            control.readyCursor()
            return
        except Exception as e:
            Q7FingerPrint.Unlock()
            txt = """Aborted: %s""" % e
            self._data = (None, (0, txt, ''))
            self.datacompleted.emit(self._data)
            control.readyCursor()
            return
        kw['isfile'] = True
        Q7FingerPrint.Unlock()
        self._data = (Q7FingerPrint(control, filedir, filename,
                                    tree, links, paths, **kw),)
        self.datacompleted.emit(self._data)


# -----------------------------------------------------------------
class Q7Window(QWidget, object):
    VIEW_CONTROL = 'C'
    VIEW_DIAG = 'D'
    VIEW_TOOLS = 'X'
    VIEW_TREE = 'T'
    VIEW_OPTION = 'O'
    VIEW_VTK = 'G'
    VIEW_FORM = 'F'
    VIEW_QUERY = 'Q'
    VIEW_SELECT = 'S'
    VIEW_PATTERN = 'P'
    VIEW_INFO = 'I'
    VIEW_LINK = 'L'
    VIEW_DIFF = 'Z'
    VIEW_MESSAGE = 'M'
    HISTORYLASTKEY = '///LAST///'
    (I_MOD_SAV, I_UMOD_SAV, I_MOD_USAV, I_UMOD_USAV, I_TREE, I_VTK, I_QUERY, I_DIFF,
     I_FORM, I_SELECT, I_PATTERN, I_DIAG, I_TOOLS, I_LINK, I_D_INF, I_D_ERR, I_L_OKL,
     I_L_NFL, I_L_NRL, I_L_NNL, I_L_IGN, I_L_ERL, I_C_SFL, I_C_SWR, I_EMPTY,
     I_MARK) = range(26)
    ICONMAPPING = {
        I_MOD_SAV: ":/images/icons/save-S-M.png",
        I_UMOD_SAV: ":/images/icons/save-S-UM.png",
        I_MOD_USAV: ":/images/icons/save-US-M.png",
        I_UMOD_USAV: ":/images/icons/save-US-UM.png",
        I_TREE: ":/images/icons/tree-load.png",
        I_VTK: ":/images/icons/vtkview.png",
        I_QUERY: ":/images/icons/operate-execute.png",
        I_DIFF: ":/images/icons/merge.png",
        I_FORM: ":/images/icons/form.png",
        I_SELECT: ":/images/icons/mark-node.png",
        I_PATTERN: ":/images/icons/pattern.png",
        I_DIAG: ":/images/icons/check-all.png",
        I_TOOLS: ":/images/icons/toolsview.png",
        I_LINK: ":/images/icons/link-node.png",
        I_D_INF: ":/images/icons/subtree-sids-warning.png",
        I_D_ERR: ":/images/icons/subtree-sids-failed.png",
        I_L_OKL: ":/images/icons/link.png",
        I_L_NFL: ":/images/icons/link-nofile.png",
        I_L_NRL: ":/images/icons/link-noreadable.png",
        I_L_NNL: ":/images/icons/link-nonode.png",
        I_L_IGN: ":/images/icons/link-error.png",
        I_L_ERL: ":/images/icons/link-ignore.png",
        I_C_SFL: ":/images/icons/check-fail.png",
        I_C_SWR: ":/images/icons/check-warn.png",
        I_EMPTY: ":/images/icons/empty.png",
        I_MARK: ":/images/icons/mark-node.png",
    }
    control_log = None
    _icons = {}
    _title = {}

    @classmethod
    def initClassConstants(cls):
        for ik in cls.ICONMAPPING:
            cls._icons[ik] = QIcon(QPixmap(cls.ICONMAPPING[ik]))
        cls._title[cls.VIEW_CONTROL] = 'Control'
        cls._title[cls.VIEW_OPTION] = 'Options'
        cls._title[cls.VIEW_TOOLS] = 'Tools'
        cls._busyx = QCursor(QPixmap(":/images/icons/cgSpy.png"))

    @classmethod
    def IC(cls, tag):
        return cls._icons[tag]

    def __init__(self, vtype, control, path, fgprintindex):
        if fgprintindex is None:
            fgprintindex = -1
        if Q7Window._icons == {}:
            Q7Window.initClassConstants()
        QWidget.__init__(self, None)
        self._stylesheet = None
        if vtype == Q7Window.VIEW_TREE:
            self._stylesheet = Q7TREEVIEWSTYLESHEET
        if vtype == Q7Window.VIEW_CONTROL:
            self._stylesheet = Q7CONTROLVIEWSTYLESHEET
        self.setupUi(self)
        # self.setAttribute(Qt.WA_DeleteOnClose)
        if self._stylesheet is not None:
            self.setStyleSheet(self._stylesheet)
        self.getOptions()
        self._timercount = 0
        self._vtype = vtype
        self._path = path
        self._control = control
        self._application = None
        self._fgindex = fgprintindex
        self._index = self.addChildWindow()
        self._lockableWidgets = []
        self._lockedWidgets = []
        fn = ''
        if self._fgindex != -1:
            fn = Q7FingerPrint.getByIndex(self._fgindex).filename
        if self._index != 0:
            tit = "%s: %s%.3d" % (OCTXT._ToolName, self._vtype, self._index)
        else:
            if vtype in self._title:
                tit = "%s: %s" % (OCTXT._ToolName, self._title[vtype])
            else:
                tit = "%s: ---" % OCTXT._ToolName
        self.setWindowTitle(tit)
        try:
            self.bBackControl.clicked.connect(self.backcontrol)
        except AttributeError:
            pass
        self._readonly = False
        if not control.verbose:
            sys.stdout = self.control_log
            sys.stderr = self.control_log

    @property
    def FG(self):
        return Q7FingerPrint.getByIndex(self._fgindex)

    def validateOption(self, name, value):
        # if (name[0]=='_'): return False
        return True

    def getOptions(self):
        try:
            if self._options is None:
                self._options = {}
        except AttributeError:
            self._options = {}
        user_options = OCTXT._readOptions(self)
        for k in dir(OCTXT):
            self.setOptionValue(k, getattr(OCTXT, k))
        for k in OCTXT._Default_Fonts:
            self.setOptionValue(k, OCTXT._Default_Fonts[k])
        if user_options is not None:
            for k in user_options:
                self.setOptionValue(k, user_options[k])
        return self._options

    def setOptions(self):
        OCTXT._writeOptions(self)

    def getOptionValue(self, name):
        return self._options[name]

    def setOptionValue(self, name, value):
        if self.validateOption(name, value):
            self._options[name] = value
            return value
        return self._options[name]

    def matchFileExtensions(self, filename):
        xl = self.getFileExtensions()
        (name, ext) = os.path.splitext(os.path.basename(filename))
        return ((xl == []) or (ext in xl))

    def getFileExtensions(self):
        xlist = []
        if OCTXT.FilterCGNSFiles:
            xlist += OCTXT.CGNSFileExtension
        if OCTXT.FilterHDFFiles:
            xlist += OCTXT.HDFFileExtension
        if OCTXT.FilterOwnFiles:
            xlist += OCTXT.OwnFileExtension
        return xlist

    def getHistoryLastKey(self):
        return Q7Window.HISTORYLASTKEY

    def getHistory(self, fromfile=True):
        if fromfile:
            self._history = OCTXT._readHistory(self)
            if self._history is None:
                self._history = {}
        return self._history

    def getHistoryFile(self, filename):
        if self._history is None:
            return None
        for d in self._history:
            if filename in self._history[d]:
                return (d, filename)
        return None

    def destroyHistory(self):
        self._history = {}
        OCTXT._writeHistory(self)
        return self._history

    def getDirNotFoundFromHistory(self):
        nf = []
        for d in list(self._history):
            if not os.path.exists(d) and not (d == Q7Window.HISTORYLASTKEY):
                nf.append(d)
        return nf

    def getDirNoHDFFromHistory(self):
        nf = []
        for d in list(self._history):
            if not os.path.exists(d) or (d == Q7Window.HISTORYLASTKEY):
                pass
            else:
                fl = [f for f in os.listdir(d) if self.matchFileExtensions(f)]
                if not fl:
                    nf.append(d)
        return nf

    def removeDirFromHistory(self, filedir):
        if filedir in self._history:
            del self._history[filedir]
        if ((self.getLastFile() is not None) and
                (self._history[Q7Window.HISTORYLASTKEY][0] == filedir)):
            del self._history[Q7Window.HISTORYLASTKEY]
        OCTXT._writeHistory(self)

    def removeFileFromHistory(self, filedir, filename):
        if (filedir in self._history and
                (filename in self._history[filedir])):
            self._history[filedir].remove(filename)
        if ((self.getLastFile() is not None) and
                (self._history[Q7Window.HISTORYLASTKEY][1] == filename)):
            del self._history[Q7Window.HISTORYLASTKEY]
        OCTXT._writeHistory(self)

    def setHistory(self, filedir, filename):
        for d in list(self._history):
            if d == filedir:
                if filename not in self._history[filedir]:
                    self._history[filedir].append(filename)
            else:
                self._history[filedir] = [filename]
        if self._history == {}:
            self._history[filedir] = [filename]
        self._history[Q7Window.HISTORYLASTKEY] = (filedir, filename)
        OCTXT._writeHistory(self)
        return self._history

    def getQueries(self):
        self._queries = OCTXT._readQueries(self)
        if self._queries is None:
            self._queries = []
        return self._queries

    def setQueries(self):
        OCTXT._writeQueries(self)
        return self._queries

    def getLastFile(self):
        if ((self._history == {}) or
                    Q7Window.HISTORYLASTKEY not in self._history):
            return None
        return self._history[Q7Window.HISTORYLASTKEY]

    def addChildWindow(self):
        if self.FG is None:
            return 0
        self._index = self.FG.addChild(self._vtype, self)
        l = [self.FG._status, self._vtype, '%.3d' % self._index]
        l += [self.FG.filedir, self.FG.filename, self._path]
        self._control.addLine(l, self.FG)
        return self._index

    def closeEvent(self, event):
        self._control.delLine('%.3d' % self._index)
        if self.FG is not None:
            self.FG.closeView(self._index)
        event.accept()

    def backcontrol(self):
        if self.FG is not None:
            self.FG.raiseControlView()
        self._control.selectLine('%.3d' % self._index)

    def busyCursor(self):
        self._control.lockView(True)
        Q7FingerPrint.disableAllViewsButtons(True)

    def readyCursor(self):
        self._control.lockView(False)
        Q7FingerPrint.disableAllViewsButtons(False)

    def lockView(self, lock):
        for wid in self._lockableWidgets:
            if not lock and wid in self._lockedWidgets:
                wid.setDisabled(lock)
                self._lockedWidgets.remove(wid)
            if lock and wid.isEnabled():
                self._lockedWidgets.append(wid)
                wid.setDisabled(lock)
        if self.FG is not None:
            self.FG._locked = lock

    def isLocked(self):
        if self.FG is None:
            return self.FG._locked
        return False

    def lockable(self, widget):
        self._lockableWidgets.append(widget)

    def refreshScreen(self):
        if self._control._application is not None:
            self._control._application.processEvents()

    def setLabel(self, it, text):
        it.setText(text)
        it.setFont(QFont("Courier"))
        it.setReadOnly(True)

    def _T(self, msg):
        if self.getOptionValue('NAVTrace'):
            print('### cg_look:', msg)


# -----------------------------------------------------------------
class Q7FingerPrint:
    STATUS_UNCHANGED = 'U'
    STATUS_MODIFIED = 'M'
    STATUS_SAVEABLE = 'S'
    STATUS_CONVERTED = 'C'
    STATUS_LIST = (STATUS_UNCHANGED, STATUS_MODIFIED, STATUS_CONVERTED,
                   STATUS_SAVEABLE)
    __mutex = QMutex()
    __chloneproxy = None
    __viewscounter = 0
    __fingerprintcounter = 0
    __extension = []  # the single variable that holds the fingerprint reference

    @classmethod
    def proxy(cls):
        cls.Lock()
        if cls.__chloneproxy is None:
            thrd = OCTXT.ActivateMultiThreading
            # BLOCK ACTUAL MULTI-THREADING HERE
            cls.__chloneproxy = Q7CHLoneProxy(thrd)
        cls.Unlock()
        return cls.__chloneproxy

    @classmethod
    def killProxy(cls):
        cls.Lock()
        if cls.__chloneproxy is not None:
            cls.__chloneproxy.kill()
        cls.Unlock()

    @classmethod
    def Lock(cls):
        cls.__mutex.lock()

    @classmethod
    def Unlock(cls):
        cls.__mutex.unlock()

    @classmethod
    def fileconversion(cls, fdir, filein, control):
        fileout = OCTXT.TemporaryDirectory + '/' + filein + '.hdf'
        count = 1
        while os.path.exists(fileout):
            fileout = OCTXT.TemporaryDirectory + '/' + filein + '.%.3d.hdf' % count
            count += 1
        com = '(cd %s; %s -f -h %s %s)' % (fdir,
                                           OCTXT.ADFConversionCom,
                                           filein, fileout)
        os.system(com)
        return fileout

    @classmethod
    def treeLoad(cls, control, selectedfile):
        control.loadOptions()
        proxy = cls.proxy()
        proxy.load(control, selectedfile)
        cls.refreshScreen()
        return

    @classmethod
    def treeSave(cls, control, fgprint, f, saveas):
        flags = CGNS.MAP.S2P_DEFAULT
        if OCTXT.CHLoneTrace:
            flags |= CGNS.MAP.S2P_TRACE
        if not saveas:
            flags |= CGNS.MAP.S2P_UPDATE
            flags |= CGNS.MAP.S2P_DELETEMISSING
        tree = fgprint.tree
        lazylist = [CGU.getPathNoRoot(path) for path in list(fgprint.lazy)]
        try:
            CGNS.MAP.save(f, tree, links=fgprint.links, flags=flags, skip=lazylist)
        except (CGNS.MAP.error,) as chlex:
            txt = """The current save operation has been aborted (CHLone):"""
            control.readyCursor()
            MSG.wError(control, 130, txt, chlex.args[0][1])
            return None
        except (Exception,) as e:
            txt = """The current save operation has been aborted: %s""" % e
            control.readyCursor()
            MSG.wError(control, 131, txt, '')
            return None
        fgprint.updateFileStats(f, saveas=True)

    @classmethod
    def closeAllTrees(cls):
        while cls.__extension:
            x = cls.__extension[0]
            x.closeAllViews()

    @classmethod
    def refreshScreen(cls):
        QCoreApplication.processEvents()

    @classmethod
    def disableAllViewsButtons(cls, lock):
        cls.Lock()
        for x in cls.__extension:
            for vtype in x.views:
                for (v, i) in x.views[vtype]:
                    for wid in v._lockableWidgets:
                        if not lock and wid in v._lockedWidgets:
                            wid.setDisabled(lock)
                            v._lockedWidgets.remove(wid)
                        if lock and wid.isEnabled():
                            v._lockedWidgets.append(wid)
                            wid.setDisabled(lock)
                    if v.FG is not None:
                        v.FG._locked = lock
        cls.Unlock()

    @classmethod
    def raiseView(cls, idx):
        cls.Lock()
        for x in cls.__extension:
            for vtype in x.views:
                for (v, i) in x.views[vtype]:
                    if i == int(idx):
                        v.raise_()
        cls.Unlock()

    @classmethod
    def infoView(cls, idx):
        f = cls.getFingerPrint(idx)
        v = cls.getView(idx)
        if f is None:
            return (None, None, None)
        if not f.isfile:
            return (f, None, None)
        return (f, v, f.getInfo())

    @classmethod
    def getView(cls, idx):
        for x in cls.__extension:
            for vtype in x.views:
                for (v, i) in x.views[vtype]:
                    if i == int(idx):
                        return v
        return None

    @classmethod
    def getViewType(cls, idx):
        vw = cls.getView(idx)
        for x in cls.__extension:
            for vtype in x.views:
                for (v, i) in x.views[vtype]:
                    if (v == vw) and (i == int(idx)):
                        return vtype
        return None

    @classmethod
    def getFingerPrint(cls, idx):
        ix = int(idx)
        for x in cls.__extension:
            for vtype in x.views:
                for (v, i) in x.views[vtype]:
                    if i == ix:
                        return x
        return None

    @classmethod
    def getUniqueTreeViewIdList(cls):
        r = set()
        for x in cls.__extension:
            for vtype in x.views:
                if vtype == Q7Window.VIEW_TREE:
                    for (v, i) in x.views[vtype]:
                        r.add(i)
        return list(r)

    @classmethod
    def getExpandedFilenameList(cls):
        l = []
        for x in cls.__extension:
            l.append("%s/%s" % (x.filedir, x.filename))
        return l

    @classmethod
    def removeNoMoreView(cls):
        for fg in list(Q7FingerPrint.__extension):
            if fg.views == {}:
                fg.model.doRelease()
                fg.doRelease()
                Q7FingerPrint.__extension.remove(fg)

    @classmethod
    def getByIndex(cls, index):
        for fg in Q7FingerPrint.__extension:
            if fg.index == index:
                return fg
        return None

    # -------------------------------------------------------------
    def __init__(self, control, filedir, filename, tree, links, paths, **kw):
        if control is None:
            return  # __root instance, empty
        self.index = Q7FingerPrint.__fingerprintcounter
        self.filename = filename
        self.tree = tree
        self.filedir = filedir
        self.links = links
        self.model = None
        self.depth = 0
        self.nodes = 0
        self.views = {}
        self.control = control
        self.converted = False
        self.isfile = False
        self.infoData = {}
        self.tmpfile = ''
        self._kw = kw
        self._status = []
        self._locked = False
        if 'isfile' in kw:
            self.isfile = True
        if 'converted' in kw:
            self.converted = kw['converted']
            self.tmpfile = kw['convertedAs']
            if self.converted:
                self._status = []
        self.lazy = {}
        for p in paths:
            self.lazy['/CGNSTree' + p[0]] = p[1]
        if ((self.lazy == {}) and
            (checkFilePermission(filedir + '/' + filename, write=True))):
            self._status = [Q7FingerPrint.STATUS_SAVEABLE]
        Q7FingerPrint.__extension.append(self)
        Q7FingerPrint.__fingerprintcounter += 1
        self.updateFileStats(filedir + '/' + filename)

    def __len__(self):
        return 0

    def isLocked(self):
        return self._locked

    def updateNodeData(self, pathdict):
        tfile = "%s/%s" % (self.filedir, self.filename)
        slp = OCTXT.LinkSearchPathList
        slp += [self.filedir]
        minpath = CGU.getPathListCommonAncestor(list(pathdict))
        flags = CGNS.MAP.S2P_DEFAULTS
        if OCTXT.CHLoneTrace:
            flags |= CGNS.MAP.S2P_TRACE
        if OCTXT.FollowLinksAtLoad:
            flags |= CGNS.MAP.S2P_FOLLOWLINKS
        (t, l, p) = CGNS.MAP.load(tfile, flags=flags, path=minpath, lksearch=slp,
                                  update=pathdict)
        return (t, l, p)

    def updateFileStats(self, fname, saveas=False):
        (filedir, filename) = (os.path.normpath(os.path.dirname(fname)),
                               os.path.basename(fname))
        self.filename = filename
        self.filedir = filedir
        self.version = CGU.getVersion(self.tree)
        self.removeTreeStatus(Q7FingerPrint.STATUS_MODIFIED)
        self.addTreeStatus(Q7FingerPrint.STATUS_UNCHANGED)
        if saveas:
            self.converted = False
            self.isfile = True
            self.tmpfile = ''
            self.removeTreeStatus(Q7FingerPrint.STATUS_CONVERTED)
            self.addTreeStatus(Q7FingerPrint.STATUS_SAVEABLE)

    def isFile(self):
        return self.isfile

    def isLink(self, path):
        pth = CGU.getPathNoRoot(path)
        for lk in self.links:
            if lk[3] == pth:
                return lk
        return False

    def fileHasChanged(self):
        dnow = {}
        self.readInfoFromOS(dnow)
        for k in dnow:
            if dnow[k] != self.infoData[k]:
                print(k, dnow[k], self.infoData[k])
                return True
        return False

    def readInfoFromOS(self, d):
        f = '%s/%s' % (self.filedir, self.filename)
        d['eFilename'] = f
        d['eDirSource'] = self.filedir
        d['eFileSource'] = self.filename
        d['eTmpFile'] = self.tmpfile
        d['eDepth'] = self.depth
        d['eNodes'] = self.nodes
        d['eVersion'] = str(self.version)
        d['eVersionHDF5'] = '???'
        try:
            st = os.stat(f)
            d['eFileSize'] = "%.3f Mb" % (1.0 * st[6] / (1024 * 1024))
            d['eMergeSize'] = "%.3f Mb" % (1.0 * st[6] / (1024 * 1024))
            dfmt = "%Y-%m-%d %H:%M"
            d['eLastDate'] = time.strftime(dfmt, time.localtime(int(st[7])))
            d['eModifDate'] = time.strftime(dfmt, time.localtime(int(st[8])))
            e = getpwuid(st[4])
            g = getgrgid(st[5])
            d['eOwner'] = e[0]
            d['eGroup'] = g[0]
            d['cNoFollow'] = False
            d['cHasLinks'] = (len(self.links) != 0)
            d['cSameFS'] = False
            d['cBadLinks'] = False
            d['cModeProp'] = False
            m = ""
            if st[0] & stat.S_IRUSR:
                m += "r"
            else:
                m += "-"
            if st[0] & stat.S_IWUSR:
                m += "w"
            else:
                m += "-"
            if st[0] & stat.S_IXUSR:
                m += "x"
            else:
                m += "-"
            if st[0] & stat.S_IRGRP:
                m += "r"
            else:
                m += "-"
            if st[0] & stat.S_IWGRP:
                m += "w"
            else:
                m += "-"
            if st[0] & stat.S_IXGRP:
                m += "x"
            else:
                m += "-"
            if st[0] & stat.S_IROTH:
                m += "r"
            else:
                m += "-"
            if st[0] & stat.S_IWOTH:
                m += "w"
            else:
                m += "-"
            if st[0] & stat.S_IXOTH:
                m += "x"
            else:
                m += "-"
            d['eRights'] = m
        except OSError:
            pass
        d['cConverted'] = self.converted
        d['cADF'] = self.converted
        d['cHDF5'] = not self.converted
        d['cREAD'] = not self.isSaveable()
        d['cMODIFY'] = self.isSaveable()
        d['cNODATA'] = False
        d['cHasInt64'] = False
        return d

    def getInfo(self, force=False):
        if force or not self.infoData:
            if self.isFile:
                self.readInfoFromOS(self.infoData)
            else:
                print('MEMORY TREE')
        return self.infoData

    def raiseControlView(self):
        self.control.show()
        self.control.raise_()

    def addChild(self, viewtype, view):
        Q7FingerPrint.__viewscounter += 1
        idx = Q7FingerPrint.__viewscounter
        if viewtype not in self.views:
            self.views[viewtype] = []
        self.views[viewtype].append((view, idx))
        return Q7FingerPrint.__viewscounter

    def closeView(self, i):
        idx = int(i)
        fg = self.getFingerPrint(idx)
        vw = self.getView(idx)
        if vw is not None:
            vt = self.getViewType(idx)
            if vt in self.views:
                self.views[vt].remove((vw, idx))
                if self.views[vt] == []:
                    del self.views[vt]
            vw.doRelease()
            vw.close()
        Q7FingerPrint.removeNoMoreView()

    def doRelease(self):
        self.model = None
        self.tree = None
        self.links = None

    def unlockAllViews(self):
        self.lockAllViews(lock=False)

    def lockAllViews(self, lock=True):
        vtlist = list(self.views)
        for vtype in vtlist:
            for (v, i) in self.views[vtype]:
                v.lockView(lock)

    def closeAllViews(self):
        vtlist = list(self.views)
        for vtype in vtlist:
            for (v, i) in self.views[vtype]:
                self.closeView(i)

    def isModified(self):
        return (Q7FingerPrint.STATUS_MODIFIED in self._status)

    def isSaveable(self):
        return (Q7FingerPrint.STATUS_SAVEABLE in self._status)

    def removeTreeStatus(self, status):
        if status not in Q7FingerPrint.STATUS_LIST:
            return
        if status in self._status:
            self._status.remove(status)
        self.control.updateViews()

    def addTreeStatus(self, status):
        if status not in Q7FingerPrint.STATUS_LIST:
            return
        if status not in self._status:
            self._status.append(status)
        self.control.updateViews()

    def pushGrammarPaths(self):
        if not OCTXT.ValKeyList:
            return
        tag = OCTXT.ValKeyList[0]
        pths = []
        self._oldsys = sys.path
        for p in OCTXT.GrammarSearchPathList:
            pths.append(str(p))
        for p in sys.path:
            pths.append(str(p))
        pths_uniq = []
        for p in pths:
            if p not in pths_uniq:
                pths_uniq.append(p)
        sys.path = pths_uniq

    def popGrammarPaths(self):
        sys.path = self._oldsys

    def nextGrammarTag(self):
        keyset = set(OCTXT.ValKeyList)
        keyset.add('SIDS')
        for t in keyset:
            yield t

# -----------------------------------------------------------------



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
import CGNS.PAT.cgnsutils    as CGU


def mergeTrees(treelist):
    """Merge a list of CGNS/Python trees into a single one
    Actual trees are modified
    List order is meaningful: the first tree is set as the master tree,
    if a conflict occurs and the FORCEFIRST flag is on, the master tree
    overwrites the slave tree...
    There is no loop detection"""
    if (len(treelist) < 2): return treelist[0]
    tmaster = treelist[0]
    for tslave in treelist[1:]:
        tmaster = mergeTwoTrees(tmaster, tslave)
    return tmaster


def mergeTwoTrees(TM, TS, path=''):
    TR = [TM[0], TM[1], [], TM[3]]
    C_M = set(CGU.childrenNames(TM))
    C_S = set(CGU.childrenNames(TS))
    C_L = C_M.symmetric_difference(C_S)
    for C in C_L:
        # print 'add   ',path+'/'+C
        if (C in C_M):
            T = TM
        else:
            T = TS
        C_N = CGU.hasChildNode(T, C)
        TR[2].append(C_N)
    C_L = C_M.intersection(C_S)
    for C in C_L:
        # print 'merge ',path+'/'+C
        TR[2].append(mergeTwoTrees(CGU.hasChildNode(TM, C),
                                   CGU.hasChildNode(TS, C),
                                   path + '/' + C))
    return TR

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
import CGNS.PAT.cgnskeywords as CK
import CGNS.PAT.cgnstypes    as CT

stringsphinx0 = """

----

.. _X%s:

%s
%s

 * Name

   * %s
 """

stringsphinx1 = """

 * Data-Type: %s

 * Dimensions/DimensionValues

 * Cardinality: %s

"""

stringsphinx2 = """
 * Child Nodes

   - :ref:`%s <X%s>` (%s)\n"""

stringsphinx3 = """   - :ref:`%s <X%s>` (%s)\n"""

stringsphinx4 = """   * %s\n"""


# a ref is :ref:`XCGNSBase_t`
def showBoxes():
    s = ""
    lt = list(CT.types)
    lt.sort()
    for t in lt:
        st = "-" * len(t)
        s += stringsphinx0 % (t, t, st, CT.types[t].names[0])
        if (len(CT.types[t].names) > 1):
            for n in CT.types[t].names[1:]:
                s += stringsphinx4 % n
        datatype = ""
        for d in CT.types[t].datatype:
            datatype += "%s " % d
        cardinality = CT.types[t].cardinality
        s += stringsphinx1 % (datatype, cardinality)
        if (CT.types[t].children):
            s += stringsphinx2 % (CT.types[t].children[0][0],
                                  CT.types[t].children[0][0],
                                  CT.types[t].children[0][1])
        if (CT.types[t].children > 1):
            for c in CT.types[t].children[1:]:
                s += stringsphinx3 % (c[0], c[0], c[1])
    return s


def generateSphinx(file):
    f = open(file, 'w+')
    f.write(showBoxes())
    f.close()

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS.VAL - Python package for CFD General Notation System - VALidater
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

TESTS = []

#  -------------------------------------------------------------------------
tag = 'gridconnectivity'
diag = True
vertexsize = 20
cellsize = 7
ntris = 12


def makeCorrectTree():
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, 'Base', 3, 3)
    z1 = CGL.newZone(b, 'Zone1', NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order='F'))
    g = CGL.newGridCoordinates(z1, 'GridCoordinates')
    d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype='float64', order='F'))
    s = NPY.array([[vertexsize, cellsize, 0]], dtype='int32', order='F')
    z2 = CGL.newZone(b, 'Zone2', s, CGK.Unstructured_s)
    g = CGL.newGridCoordinates(z2, 'GridCoordinates')
    d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((vertexsize), dtype='float64', order='F'))
    tetras = CGL.newElements(z2, 'TETRAS', CGK.TETRA_4_s,
                             NPY.array([[1, cellsize]], 'i', order='F'),
                             NPY.ones((cellsize * 4), dtype='int32'))
    tris = CGL.newElements(z2, 'TRIS', CGK.TRI_3_s,
                           NPY.array([[cellsize + 1, cellsize + ntris]], 'i', order='F'),
                           NPY.ones((ntris * 3), dtype='int32'))
    z3 = CGU.copyNode(z1, 'Zone3')
    b[2].append(z3)
    z4 = CGU.copyNode(z2, 'Zone4')
    b[2].append(z4)
    z = [z1, z2, z3, z4]
    return (T, b, z)


(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_3', 'Zone3', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 1], [1, 7], [1, 9]], order='F'))
CGL.newPointRange(gc, name=CGK.PointRangeDonor_s, value=NPY.array([[1, 1], [1, 7], [1, 9]], order='F'))
gc = CGL.newGridConnectivity(zgc, 'join1_2', 'Zone2', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order='F'))
CGL.newIndexArray(gc, CGK.PointListDonor_s, value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order='F'))
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'gridconnectivity bad datatype'
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_2', 'Zone2', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order='F'))
CGL.newIndexArray(gc, CGK.PointListDonor_s, value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order='F'))
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
gc[1] = NPY.array([1], order='F')
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'gridconnectivity unfound zonedonor'
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_2', 'Zone6', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order='F'))
CGL.newIndexArray(gc, CGK.PointListDonor_s, value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order='F'))
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'gridconnectivity bad pointrange/list'
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_2', 'Zone2', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order='F'))
CGL.newIndexArray(gc, CGK.PointList_s, value=NPY.array([[1, 5], [1, 1], [1, 9]], order='F'))
CGL.newIndexArray(gc, CGK.PointListDonor_s, value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order='F'))
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
gc = CGL.newGridConnectivity(zgc, 'join1_3', 'Zone3', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, name=CGK.PointRangeDonor_s, value=NPY.array([[1, 1], [1, 7], [1, 9]], order='F'))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'gridconnectivity bad gridlocation'
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_2a', 'Zone2', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 4], [1, 1], [1, 8]], order='F'))
CGL.newIndexArray(gc, CGK.PointListDonor_s, value=NPY.array([range(1, cellsize)], order='F'))
CGL.newGridLocation(gc, value=CGK.CellCenter_s)
gc = CGL.newGridConnectivity(zgc, 'join1_2b', 'Zone2', ctype=CGK.Overset_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 7], [1, 1]], order='F'))
CGL.newIndexArray(gc, CGK.PointListDonor_s, value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order='F'))
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'gridconnectivity bad range on zonedonor #1'
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_2', 'Zone2', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order='F'))
CGL.newIndexArray(gc, CGK.PointListDonor_s, value=NPY.array([range(cellsize - 1, cellsize + ntris + 1)], order='F'))
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'gridconnectivity bad range on zonedonor #2'
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_2', 'Zone2', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order='F'))
CGL.newIndexArray(gc, CGK.PointListDonor_s, value=NPY.array([range(cellsize + 2, cellsize + ntris + 2)], order='F'))
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'gridconnectivity bad range on zonedonor #3'
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_3', 'Zone3', ctype=CGK.Overset_s)
CGL.newPointRange(gc, value=NPY.array([[1, 1], [1, 7], [1, 9]], order='F'))
CGL.newPointRange(gc, name=CGK.PointRangeDonor_s, value=NPY.array([[1, 1], [1, 7], [1, 9]], order='F'))
CGL.newGridLocation(gc, value=CGK.CellCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = 'gridconnectivity abutting not a face'
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, 'join1_3', 'Zone3', ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 7], [1, 9]], order='F'))
CGL.newPointRange(gc, name=CGK.PointRangeDonor_s, value=NPY.array([[1, 5], [1, 7], [1, 9]], order='F'))
TESTS.append((tag, T, diag))



#!/usr/bin/env python

# [ name, value, children, type ]
import CGNS.MAP
import CGNS.PAT
import CGNS.PAT.cgnsutils

(tree, links) = CGNS.MAP.load("./001Disk.hdf", CGNS.MAP.S2P_FOLLOWLINKS | CGNS.MAP.S2P_TRACE)

ref = ['ReferenceState', None, [], CGNS.PAT.ReferenceState_t]
ref[2].append(['Density', dens, [], 'DataArray_t'])

node = CGNS.PAT.cgnsutils.getNodeByPath(tree, "/Disk")
node[2].append(ref)

CGNS.MAP.save("./001Disk-new.hdf", CGNS.MAP.S2P_FOLLOWLINKS | CGNS.MAP.S2P_TRACE, tree, links)



#  -------------------------------------------------------------------------
#  pyCGNS.NAV - Python package for CFD General Notation System - NAVigater
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

from qtpy.QtWidgets import QWidget
from qtpy.QtGui import QTextCursor

from CGNS.NAV.Q7HelpWindow import Ui_Q7HelpWindow


# -----------------------------------------------------------------
class Q7Help(QWidget, Ui_Q7HelpWindow):
    def __init__(self, control, key=None, doc=None):
        super(Q7Help, self).__init__()
        self.setupUi(self)
        self._control = control
        if (doc is None) and (key not in HELPDOCS):
            return
        self.eHelp.setAcceptRichText(True)
        self.eHelp.clear()
        if key is not None:
            self.setWindowTitle("CGNS.NAV: Help on " + HELPDOCS[key][0])
            self.eHelp.insertHtml(HELPDOCS[key][1])
        elif doc is not None:
            self.setWindowTitle("CGNS.NAV: Contextual Documentation")
            self.eHelp.insertHtml(doc)
        else:
            return
        self.eHelp.moveCursor(QTextCursor.Start, QTextCursor.MoveAnchor)
        self.eHelp.ensureCursorVisible()
        self.eHelp.setReadOnly(True)

    def close(self):
        self._control.help = None
        QWidget.close(self)


HELPDOCS = {
    # --------------------------------------------------------------------------
    'Control': ('Control panel',
                """
                <h2>Control panel</h2>
                The main panel of CGNS.NAV is a summary of views. The list gives you
                short information about the file/tree being viewed and allows you some
                basic operations.
                <p>
                You can open new trees, close views or close all <i>CGNS.NAV</i> from this
                panel. 
                
                <h3>Table entries</h3>
                Each line of the control panel is a view. Each view has a number, which
                counts from 001 up to the max number of views you have. The status and the
                type of the view are indicated by an icon.<i>We use the term <b>view</b>
                when a CGNS.NAV window appears in this list
                and <b>panel</b> when the window doesn't appear.
                <p>
                
                The <b>status</b> changes each time you modify the CGNS/Python tree or
                its links or any other parameter that would lead to save a different file
                than the one that was loaded at first.
                
                The <b>status</b> may indicate you cannot save the file, for example if
                the file is read only, or if the file has been converted; In this latter
                case, the actual file name is a temporary file name. The <b>Info panel</b>
                gives you more details about the file itself: right click on a row in the
                control panel list and select <b>View Information</b> to get
                the <b>Info panel</b>.
                <p>
                The <b>status</b> icons are:
                
                <table>
                <tr><td><img source=":/images/icons/save-S-UM.png"></td>
                <td>File can be saved, no modification to save now</td></tr>
                <tr><td><img source=":/images/icons/save-S-M.png"></td>
                <td>File can be saved, some modifications are not saved yet</td></tr>
                <tr><td><img source=":/images/icons/save-US-UM.png"></td>
                <td>File cannot be saved, no modification to save now. You have no write access to the file, 
                you should use <b>save as</b> button.</td></tr>
                <tr><td><img source=":/images/icons/save-US-M.png"></td>
                <td>File cannot be saved, some modifications are not saved yet. You have no write access to the file, 
                you should use <b>save as</b> button.</td></tr>
                </table>
                <p>
                The second column of a control panel line is the view type:
                
                <table>
                <tr><td><img source=":/images/icons/tree-load.png"></td>
                <td> Tree view - Open from control panel</td></tr>
                <tr><td><img source=":/images/icons/vtkview.png"></td>
                <td> VTK view - Open from tree view</td></tr>
                <tr><td><img source=":/images/icons/operate-execute.png"></td>
                <td> Query view - Open from tree view</td></tr>
                <tr><td><img source=":/images/icons/form.png"></td>
                <td> Form view - Open from tree view</td></tr>
                <tr><td><img source=":/images/icons/operate-list.png"></td>
                <td> Selection view - Open from tree view</td></tr>
                <tr><td><img source=":/images/icons/check-all.png"></td>
                <td> Diagnosis view - Open from tree view</td></tr>
                <tr><td><img source=":/images/icons/link.png"></td>
                <td> Link view - Open from tree view</td></tr>
                </table>
                
                <p>
                The remaining columns of a control panel line are the directory and the 
                filename of the CGNS/Python tree and the target node of the tree. 
                These values may be empty if not relevant. For example if you create
                a tree from scratch, the directory and the file name are generated and thus
                would require a <b>save as</b> to set these values.
                
                All <b>CGNS.NAV</b> windows have a <img source=":/images/icons/top.png">
                button that raises the control panel window. The other way, you select
                a view line in the control panel and press <i>[space]</i> to raise the
                corresponding view.
                
                You can close a single view, all views related to one tree or all the trees
                using the right click menu, on a row.
                
                <h3>Buttons</h3>
                <p>
                <table>
                <tr><td><img source=":/images/icons/tree-load-g.png"></td>
                <td> load last used file</td></tr>
                <tr><td><img source=":/images/icons/tree-load.png"></td>
                <td> load a file, open the file dialog window</td></tr>
                <tr><td><img source=":/images/icons/tree-new.png"></td>
                <td> create a new CGNS/Python tree from scratch</td></tr>
                <tr><td><img source=":/images/icons/pattern.png"></td>
                <td> open the pattern database</td></tr>
                <tr><td><img source=":/images/icons/options-view.png"></td>
                <td> open the user options panel</td></tr>
                <tr><td><img source=":/images/icons/help-view.png"></td>
                <td> help</td></tr>
                <tr><td><img source=":/images/icons/view-help.png"></td>
                <td> about CGNS.NAV</td></tr>
                <tr><td><img source=":/images/icons/close-view.png"></td>
                <td> close CGNS.NAV and all its views</td></tr>
                </table>
                
                """),
    # --------------------------------------------------------------------------
    'Tree': ('Tree view',
             """
             <h2>Tree view</h2>
             
             The main window you would use for browsing/editing a CGNS/Python tree is
             the <i>Tree view</i>.
             
             <h3>Hierarchical view</h3>
             The hierarchical view is a classical tree browser. You click just before the node name to 
             expand/collapse its children node. It supports copy/cut/paste as well
             as line editing for node name change, node sidstype change and node value change.
             <p>
             
             Bold names are user defined names, the non-bold are names found in the CGNS/SIDS.
             <p>
             You can use <i>[up][down]</i>arrows to move from a node to another, if you press <i>[control]</i> 
             together with the <i>[up][down]</i> then you move from nodes of a same level.
             Pressing <i>[space]</i> sets/unsets the selection flag on the current node.
             
             <h3>Current node and selected nodes</h3>
             
             The current node is the highlighted line. The selected nodes are the lines
             with a blue flag on. Some commands require a current node to run, for example
             the <i>Form view</i> requires it. Some other commands require a list of
             selected nodes, or example the check, or the <i>paste as child for each selected node</i>. This latter 
             command is in the right-click menu.
             
             <h3>Flags</h3>
             
             There a 4 flag columns: <b>L</b> is for link flag, <b>M</b> is mark flag,
             <b>C</b> is check flag and the last <b>U</b> is user flag.<br><br>
             
             There are two link flags, <img source=":/images/icons/link.png"> indicates
             the root node of a linked-to node. The actual node is in file used for the
             current tree, but the node value and its children are located into another
             file. Use the <img source=":/images/icons/link-view.png"> icon to open the
             link view, that shows all existing links for the current tree. Or you can move
             the mouse pointer on the link icon, the actual linked-to reference would
             be displayed.
             The second link flag is
             <img source=":/images/icons/link-child.png"> that indicates the node has a
             linked-to ancestor node (its direct parent or any other parent). Depending
             on the options you set, you may or may not have the right to change this
             child-link node name, type or value.<br><br>
             
             The mark flag is set when you run a selection, or when you mark it by
             yourself by pressing <i>[space]</i>. The list of all selected nodes, for
             example after running a large query, is displayed using the
             <img source=":/images/icons/operate-list.png"> icon.<br><br>
             
             The check flag is set when you run a check with 
             <img source=":/images/icons/check-all.png">, the diagnosis list is
             open with <img source=":/images/icons/check-list.png">.
             Nodes can have a warning <img source=":/images/icons/check-warn.png">
             or an error <img source=":/images/icons/check-fail.png"> flag. Node without
             flags are ok.<br><br>
             
             The last flag column is for the user flag. The user sets/unsets the flag
             for each node by selecting the node and pressing one of the <i>[0-9]</i> keys.
             Then, by default, the corresponding number is set as flag (for example you
             have a <img source=":/images/icons/user-5.png"> if you press <i>[5]</i>.
             The flag is an informative flag, it is not used by CGNS.NAV functions.<br><br>
             
             The snapshot below shows the flag columns with various flag settings.<br><br>
             <img source=":/images/icons/help-01.png"><br>
             
             
             <h3>Editing nodes</h3>
             You can edit node name, type and values by right-cliking on the line/column
             you want to edit.<br><br>
             
             The new name is rejected if it already exists or if its syntax is not
             compliant with CGNS/SIDS requirements.<br><br>
             
             The new SIDS type should be one of the allowed types.<br><br>
             
             The new value edit depends on its SIDS type and its data type (the *D* column).
             You may have an enumerate or a plain string. In case of plain string, you have
             to provide a new value as a python value without interpretation.<br><br>
             
             In the case you want to get rid of enumerates, for example when you want
             to document a new SIDS type for CPEX proposal, you force plain using
             <i>[insert]</i> to edit the name (*NO NAME CHECK*),
             <i>[control-insert]</i> to edit the SIDS type and
             <i>[shift-insert]</i> to edit value.
             These edit methods are not performing any check and your CGNS/Python tree
             may be not-compliant, or even impossible to actually store with CGNS/HDF5!
             
             <h3>Copy/Cut/Paste</h3>
             The <i>Tree views</i> support the copy/cut/paste of sub-trees.
             
             <h3>Top Buttons</h3>
             <p>
             <table>
             <tr><td><img source=":/images/icons/save.png"></td>
             <td> save file using the last used file name and save parameters</td></tr>
             <tr><td><img source=":/images/icons/tree-save.png"></td>
             <td> save as, as a new directory and/or file name, allows to change save
             parameters such as link management</td></tr>
             <tr><td><img source=":/images/icons/pattern-save.png"></td>
             <td> save as a pattern (not available yet)</td></tr>
             <tr><td><img source=":/images/icons/level-out.png"></td>
             <td> expand one level of children nodes</td></tr>
             <tr><td><img source=":/images/icons/level-all.png"></td>
             <td> switch expand all / unexpand all</td></tr>
             <tr><td><img source=":/images/icons/level-in.png"></td>
             <td> unexpand one level of children nodes</td></tr>
             <tr><td><img source=":/images/icons/flag-all.png"></td>
             <td> set the selection flag on all the nodes. If a node is not visible, then it is flagged as well and 
             all subsequent operation on selection list would take this invisible node as argument. Note that such an 
             invisble node is actually in the CGNS/Python tree but it is not displayed</td></tr>
             <tr><td><img source=":/images/icons/flag-revert.png"></td>
             <td> revert nodes selection flag</td></tr>
             <tr><td><img source=":/images/icons/flag-none.png"></td>
             <td> remove all selection flags</td></tr>
             <tr><td><img source=":/images/icons/operate-list.png"></td>
             <td> open the selection view, gives the list of all selected nodes</td></tr>
             <tr><td><img source=":/images/icons/check-all.png"></td>
             <td> check nodes using <b>CGNS.VAL</b> tool. If the selection list is not
             empty, then only the selected nodes are checked. If the selection list is
             empty, all nodes of the tree are checked.</td></tr>
             <tr><td><img source=":/images/icons/check-clear.png"></td>
             <td> remove all check flags</td></tr>
             <tr><td><img source=":/images/icons/check-list.png"></td>
             <td> open the diagnosis view which details the check log</td></tr>
             <tr><td><img source=":/images/icons/link-select.png"></td>
             <td> not available yet</td></tr>
             <tr><td><img source=":/images/icons/link-add.png"></td>
             <td> not available yet</td></tr>
             <tr><td><img source=":/images/icons/link-delete.png"></td>
             <td> not available yet</td></tr>
             <tr><td><img source=":/images/icons/form-open.png"></td>
             <td> open the form view, gives details on a node including its data</td></tr>
             <tr><td><img source=":/images/icons/vtk.png"></td>
             <td> open the VTK view, a 3D view on the actual mesh and its associated data.
             This requires a correct parsing of CGNS/Python tree, if the tool is not able
             to understand the data (for example if your tree is not CGNS/SIDS compliant,
             then no window is open.</td></tr>
             <tr><td><img source=":/images/icons/plot.png"></td>
             <td> open the Plot view, a 2D view on some tree data.
             Same remarks as VTK view.</td></tr>
             <tr><td><img source=":/images/icons/pattern-view.png"></td>
             <td> not available yet</td></tr>
             <tr><td><img source=":/images/icons/link-view.png"></td>
             <td> open the Link view, gives details on the links</td></tr>
             <tr><td><img source=":/images/icons/operate-view.png"></td>
             <td> open the Query view, a powerful mean to edit and run Python commands on the tree</td></tr>
             <tr><td><img source=":/images/icons/check-view.png"></td>
             <td> not available yet</td></tr>
             <tr><td><img source=":/images/icons/tools-view.png"></td>
             <td> not available yet</td></tr>
             <tr><td><img source=":/images/icons/snapshot.png"></td>
             <td> creates a bitmap file with the snapshot of the current tree view.
             See user options for directory and file used for this snapshot.</td></tr>
             </table>
             
             """),
    # --------------------------------------------------------------------------
    'VTK': ('VTK view',
            """
            <h2>VTK view</h2>
            
            <h3>Selection list</h3>
            <p>
            <table>
            <tr><td><img source=":/images/icons/selected.png"></td>
            <td>The selection list entry is selected for operations</td></tr>
            <tr><td><img source=":/images/icons/unselected.png"></td>
            <td>The selection list entry is NOT selected for operations</td></tr>
            <tr><td><img source=":/images/icons/hidden.png"></td>
            <td>The selection list entry is hidden, the data is still there but it is not
            displayed in the window.</td></tr>
            </table>
            
            <h3>Icons and buttons</h3>
            <p>
            <table>
            <tr><td><img source=":/images/icons/zoompoint.png"></td>
            <td>Switch to the zoom-window mode. You draw a window using the mouse left
            button and the zoom would fit to the defined window.</td></tr>
            <tr><td><img source=":/images/icons/zoom-actor.png"></td>
            <td>Reset the view to fit the object in the window</td></tr>
            <tr><td><img source=":/images/icons/lock-legend.png"></td>
            <td>Switch to move legend mode, used to mose the reference axis draw, the
            legend and other extra information displayed on the view.</td></tr>
            <tr><td><img source=":/images/icons/value.png"></td>
            <td>Switch to the show value mode, in the case you have actual values,
            the pick of a node would show corresponding value.</td></tr>
            <tr><td><img source=":/images/icons/colors.png"></td>
            <td>Randomly change the colors of view objects.</td></tr>
            <tr><td><img source=":/images/icons/colors-bw.png"></td>
            <td>Switch the view background to balck/white.</td></tr>
            </table>
            
            
            <h3>Displaying actual values</h3>
            <p>
            
            
            <h3>User defined camera views</h3>
            <p>
            You can the current camera and reset to this stored point of view later on.
            <table>
            <tr><td><img source=":/images/icons/camera-snap.png"></td>
            <td>Sotre an existing camera</td></tr>
            </table>
            
            
            <h3>Interactions</h3>
            
            <h4>Mouse buttons</h4>
            The <i>mouse left button</i> runs the rotation (see below). The
            <i>mouse right button</i> runs the zoom, same with the <i>mouse wheel</i>
            if you have one. The <i>mouse middle button</i>, if you have one, runs a
            translation parallel to the view plane. 
            
            <h4>Rotation</h4>
            The <i>Camera axis all</i> uses the center of the view (the center of the
            camera view) as the center of rotation.<br>
            The <i>Object axis all</i> is a rotation with the center of the object
            as rotation center.<br>
            The <i>Object axis x</i> mode forces a rotation with x as single axis, same
            for y and z.<br>
            
            <h4>Key Bindings</h4>
            Keys are not case-sensitive letters, <b>s</b> and <b>S</b> are the same.
            
            <b>s</b> Surface mode rendering (toggle with wire)<br>
            <b>w</b> Wire mode rendering (toggle with surface)<br>
            <b>q</b> Surface and wire mode rendering (toggle with wire)<br>
            <b>r</b> Reset view to fit to full object<br>
            
            <b>[space]</b> Throw a ray on the view and select all actors cuting the ray<br>
            <b>z</b> Same as [space]<br>
            <b>u</b> Unselect all<br>
            <b>p</b> Pick a point in the selected Zone<br>
            
            <h4>Tips</h4>
            
            To unselect all and clear the list of selected actors, just move the mouse on
            the graphic window background and press [space].
            """),
    # --------------------------------------------------------------------------
    'Link': ('Link view',
             """
             <h2>Link view</h2>
             
             <b>When you copy a node path you SHOULD NOT add the CGNSTree root</b>
             
             <h3>Buttons</h3>
             <p>
             
             """),
    # --------------------------------------------------------------------------
    'Form': ('Form view',
             """
             <h2>Form view</h2>
             
             This view shows all information it can find about a node. The <i>Form view</i>
             displays the node and its data using different ways and allows you some
             basic operations on this node. Each tab is a view of the same node. The
             <i>table tab</i> is a raw view of dimensions, types and shows the
             actual <i>data</i> in a spreadsheet-like widget.
             
             
             <h3>Spreadsheet</h3>
             <p>
             The data is displayed in a 2D table, you can change the horizontal/vertical
             distribution using the <i>Size</i> selector. It computes for you all
             possible combinations of horizontal and vertical indicies from the original
             data size.
             
             <h3>Links</h3>
             <p>
             All details about relationship with other files are in the <i>Link</i> tab.
             It shows you the current link if the node is actually the root node of a link,
             it also shows you the parent node link if the current node is the child of a
             linked-to node. In other words, it tells you wether the node you are looking at
             is in the top file or not.
             
             <h3>Text</h3>
             <p>
             Some node data are text, it is often more readable to look at it using a real
             text tool than a spreadsheet-like tool. The <i>Text</i> tab is available only
             for text data.
             
             """),
    # --------------------------------------------------------------------------
    'Query': ('Query view',
              """
              <h3><font color=red>WARNING</font></h3>
              The queries can modify your CGNS/Python tree in such a way that it could
              become non-compliant, inconsistant in CGNS/Python or even <font color=red>
              <b>crash</b> the CGNS.NAV application</font>
              or any of its sub-applications such as MAP or VAL.<p>
              <font color=red>Use of CGNS/Python data modification requires a very
              good level of Python and CGNS skills, use at your own risks...</font>
              
              <h2>Query view</h2>
              The query view is one of CGNS.NAV most powerful feature. It provides a
              true Python scripting access on the CGNS tree, the script you write is
              applied to each node of the whole CGNS/Python tree. You have to write the
              script using the usual Python syntax and semantics, including the required
              <b>import</b>s statements, however CGNS.NAV adds a pre-script and a post-script
              to your actual text.<br><br>
              
              The pre-script is used to defined local variable that would be useful for you.
              The variables are setting the node context, you can get/set these values as
              these are the actual node:
              
              <table>
              <tr><td><b>NODE</b></td><td>the NODE as a CGNS/Python NODE, that is a list
              with <i>[ NAME, VALUE, CHILDREN, SIDSTYPE]</i></td></tr>
              <tr><td><b>NAME</b></td><td>the NAME of the NODE, same as NODE[0]</td></tr>
              <tr><td><b>VALUE</b></td><td>the VALUE of the NODE, same as NODE[1]</td></tr>
              <tr><td><b>SIDSTYPE</b></td><td>the CGNS/SIDS type of the NODE, same
              as NODE[3]</td></tr>
              <tr><td><b>CHILDREN</b></td><td>the list of CHILDREN of the NODE,
              same as NODE[2]</td></tr>
              <tr><td><b>PARENT</b></td><td>the PARENT node of current NODE</td></tr>
              <tr><td><b>TREE</b></td><td>the complete CGNS/Python TREE</td></tr>
              <tr><td><b>LINKS</b></td><td>the list of links info</td></tr>
              <tr><td><b>SKIPS</b></td><td>the list of skips info</td></tr>
              <tr><td><b>PATH</b></td><td>the PATH to the current NODE</td></tr>
              <tr><td><b>ARGS</b></td><td>The arguments tuple you may have passed (in
              the Tree view for example). Please note this is always a tuple, even if you
              have a single argument (then use ARGS[0]). The is a special case, if your
              argument is a single string then you need not to put quotes around. For
              example, you can use <i>ZoneType</i> instead of <i>'ZoneType'</i>.</td></tr>
              <tr><td><b>RESULT</b></td><td>the output of your script, this result value
              is inserted into the global result list for all nodes. Thus, you would
              rather add a tuple containing the current PATH and the result if you
              want to find back which result matches which value (see example)</td></tr>
              </table>
              
              <h3>Buttons</h3>
              <p>
              <table>
              <tr><td><img source=":/images/icons/operate-add.png"></td>
              <td>Add a new query</td></tr>
              <tr><td><img source=":/images/icons/operate-delete.png"></td>
              <td>Remove an existing query</td></tr>
              <tr><td><img source=":/images/icons/operate-save.png"></td>
              <td>Save all queries</td></tr>
              </table>
              
              """),
    # --------------------------------------------------------------------------
    'Option': ('Option panel',
               """
               <h2>Option panel</h2>
               <font color=red>Oups! the Option panel is a MODAL window, which means
               it blocks all other windows until you answer to it. So if you want to
               browse this help you have to close the Option panel window first!</font>
               <h3>Buttons</h3>
               
               <h3>Message codes</h3>
               <table>
               <tr><td>100</td><td> * </td><td>Version and Copyright notice</td></tr>
               <tr><td>101</td><td> - </td><td>CGNS.NAV Exit double check</td></tr>
               <tr><td>110</td><td> - </td><td>Directory not found during file open</td></tr>
               <tr><td>120</td><td> - </td><td>Clear file history - current not found file</td></tr>
               <tr><td>121</td><td> - </td><td>Clear file history - all not found files</td></tr>
               <tr><td>122</td><td> - </td><td>Clear file history - all files</td></tr>
               <tr><td>130</td><td> - </td><td>Save abort</td></tr>
               <tr><td>131</td><td> - </td><td>Save abort (execption caught)</td></tr>
               <tr><td>132</td><td> - </td><td>Save warning about file overwriting issue</td></tr>
               <tr><td>200</td><td> * </td><td>Error during load (returns actual CGNS/HDF5 layer error)</td></tr>
               <tr><td>201</td><td> * </td><td>Fatal error during load</td></tr>
               <tr><td>500</td><td> - </td><td>Bad data for a VTK display</td></tr>
               <tr><td>501</td><td> - </td><td>VTK version lower than v6.0</td></tr>
               <tr><td>502</td><td> - </td><td>VTK version lower than v5.8</td></tr>
               <tr><td>302</td><td> - </td><td>Node selection required for Form view</td></tr>
               <tr><td>310</td><td> - </td><td>Create link as new node</td></tr>
               <tr><td>311</td><td> - </td><td>Remove link entry</td></tr>
               <tr><td>370</td><td> - </td><td>Leave query panel</td></tr>
               <tr><td>371</td><td> - </td><td>Apply changes to all</td></tr>
               <tr><td>400</td><td> * </td><td>Show diag view grammer used</td></tr>
               </table>
               
               <p>
               <img source=":/images/icons/unselected.png">
               """),
    # --------------------------------------------------------------------------
    'File': ('File panel',
             """
             <h2>File panel</h2>
             <font color=red>Oups! the File panel is a MODAL window, which means
             it blocks all other windows until you answer to it. So if you want to
             browse this help you have to close the File panel window first!</font>
             <p>
             The same panel is used for both <b>load</b> and <b>save</b> file selection.
             The <b>Cancel</b> button would abort the current load or save
             operation, but once the the load/save process is strated there is no way
             to abort it.
             <p>
             There are two tabs, the <i>Selection</i> tab is the default one, you select
             there the directory and filename of the target file you want to load or to
             save. The <I>Options</i> tab sets the behavior of the file filter and the
             history of your loads/saves.
             <P>
             The <i>File panel</i> filters the CGNS/HDF/ADF files depending on the options
             you set. Files can be sorted by name, size, type or modification date by
             clicking on the corresponding column header.
             <p>
             Unchecking the <i>Show directories</i> checkbox would mask all directories
             from the table, thus making it more easy to find your file in large lists.
             
             <h3>Load</h3>
             The <i>File panel</i> options checkboxes, such as <i>Do Not load large data</i>
             are used only for the current load/save. The initial values of these checkboxes,
             each time you open the panel, are get from your overall options (see the
             <img source=":/images/icons/option-view.png"> <i>Options Panel</i>).
             <p>
             <img source=":/images/icons/h_files_checks.png">
             <table>
             <tr><td><b>Do not load large data</b></td><td>Skip large arrays, the threshold
             is set in the <i>Options Panel</i></td></tr>
             <tr><td><b>Follow links</b></td><td>Recursively load files you can reach through linksSkip large arrays, 
             the threshold is
             set in the <i>Options Panel</i></td></tr>
             <tr><td><b>Open as read-only</b></td><td>Prevent any attempt to modify the
             tree. This guard is unlocked if you do a <I>Save as</i>, the tree is then
             associated to a new file and becames writable.</td></tr>
             </table>
             
             <h3>Save</h3>
             The save policy is complex as saving a CGNS tree can be an update or
             an overwrite. 
             The save checkboxes are <i>overwrite</i> and <i>delete missing</i>, the behavior
             of the same is summerized in this table:
             <table>
             <tr><td><img source=":/images/icons/w_file_checks_00.png"></td><td></td></tr>
             <tr><td><img source=":/images/icons/w_file_checks_10.png"></td><td></td></tr>
             <tr><td><img source=":/images/icons/w_file_checks_01.png"></td><td></td></tr>
             <tr><td><img source=":/images/icons/w_file_checks_11.png"></td><td></td></tr>
             </table>
             
             <h3>Buttons</h3>
             <p>
             <img source=":/images/icons/unselected.png">
             """),
    # --------------------------------------------------------------------------
    'Selection': ('Selection view',
                  """
                  <h2>Selection view</h2>
                  
                  The <i>Selection view</i> lists the selected nodes resulting from a query or
                  user performed selection. The only action you can perform on this list is
                  to change the values, datatypes or SIDS types, or to save the list.
                  
                  The actual purpose of this view is to reduce the view of a tree for local
                  modifications. For example you perform a query to retrieve a specific node
                  and then you edit the selected nodes.
                  
                  You can run the query and display the result at once using the command line,
                  for example:
                  <pre>CGNS.NAV -q '100. Family boundaries' File.cgns</pre>
                  
                  <h3>Buttons</h3>
                  <p>
                  <table>
                  <tr><td><img source=":/images/icons/control.png">
                  <img source=":/images/icons/unselected.png">
                  <img source=":/images/icons/node-sids-closed.png">
                  </td>
                  <td> unsused</td></tr>
                  <tr><td><img source=":/images/icons/select-add.png"></td>
                  <td> select all lines</td></tr>
                  <tr><td><img source=":/images/icons/falg-revert.png"></td>
                  <td> revert line selection</td></tr>
                  <tr><td><img source=":/images/icons/select-delete.png"></td>
                  <td> unselect all lines</td></tr>
                  <tr><td><img source=":/images/icons/flag-none"></td>
                  <td> remove a line from the selection line</td></tr>
                  <tr><td><img source=":/images/icons/select-save"></td>
                  <td> save the selection list in a file</td></tr>
                  </table>
                  
                  <h3>User defined edition</h3>
                  The node value editing can be a plain text edition or a combo-box enumerate.
                  In that latter case the user can define a function to map the list of
                  allowed value with respect to the node path (name for example) and the node
                  type path (that is the list of types for each node in the path to the
                  current node). The function should be a python class in a python file,
                  the file name is <pre>default.py</pre> in the directory
                  <pre>$HOME/.CGNS.NAV/funtions</pre>. There is an example of such a file:
                  <pre>
                  # -----------------------------------------------------------------
                  # Mandatory name
                  #
                  import CGNS.PAT.cgnskeywords as CGK
                  
                  class Q7UserFunction(object):
                    __attributes={
                      'artviscosity':['none','dissca','dismat','dismrt'],
                      'fluid':['rg','pg'],
                      'flux':['jameson','roe','vleer','coquel_d','coquel_i','ausmp','rbc','rbci'],
                      'ode':['backwardeuler','rk4'],
                      'phymod':['euler','nslam','nstur'],
                    }
                    # --- mandatory method
                    def getEnumerate(self,namepath,typepath):
                      if (namepath[-1] in self.__attributes):
                          return self.__attributes[namepath[-1]]
                      else:
                          return CGK.cgnsenums[typepath[-1]]
                      return None
                  
                  # --- last line
                  </pre>
                  """),
    # --------------------------------------------------------------------------
    'Diagnosis': ('Diagnosis view',
                  """
                  <h2>Diagnosis view</h2>
                  
                  A <i>Diagnosis view</i> is associated to one <i>Tree view</i> once a check
                  of its CGNS/Python tree has been performed. The <i>Diagnosis view</i>
                  shows the errors and warnings per node, filter the error/warning type you
                  want to see or suppress the warnings display. You can browse these diagnosis
                  and go back to the targeted node in the <i>Tree view</i> by pressing the
                  <i>[space]</i> key.
                  
                  <h3>Filters</h3>
                  <p>
                  You can select an error/warning code in the combo-box, it shows all the
                  error/warning codes the check has collected. In the case you have too
                  many warnings, you can just ignore them by un-setting the <i>Warnings</i>
                  checkbox.
                  
                  <h3>Buttons</h3>
                  <p>
                  <table>
                  <tr><td><img source=":/images/icons/level-in.png"></td>
                  <td> expand all errors/warnings</td></tr>
                  <tr><td><img source=":/images/icons/level-out.png"></td>
                  <td> collapse all errors/warnings</td></tr>
                  <tr><td><img source=":/images/icons/node-sids-opened.png"></td>
                  <td> go to previous filtered error/warning</td></tr>
                  <tr><td><img source=":/images/icons/selected.png"></td>
                  <td> go to next filtered error/warning</td></tr>
                  <tr><td><img source=":/images/icons/check-save.png"></td>
                  <td> save errors/warnings as a Python importable file (see hereafter)</td></tr>
                  </table>
                  
                  <h3>Save diagnosis</h3>
                  <p>
                  You can save all the diagnosis in a file. The file contains a Python
                  dictionnary in the variable <i>data</i> with the node paths as key.
                  For a key, the value is the list of diagnosis, each diagnosis is a tuple
                  of three strings: the error/warning code, the level, the message.
                  Here is an example of such a file:
                  <pre>
                  data={
                  '/SquaredNozzle/INJ3/.Solver#Trigger/next_iteration':
                    ("S004","E","DataType [I4] not allowed for this node"),
                  '/SquaredNozzle/INJ3/.Solver#Trigger/next_state':
                    ("S004","E","DataType [I4] not allowed for this node")
                  }
                  </pre>
                  
                  """),
    # --------------------------------------------------------------------------
    'Info': ('Info panel',
             """
             <h2>Info panel</h2>
             
             Gives all details on the top file you use to load/save the target CGNS/Tree.
             
             
             <b>In case of links, only the top file is detailled.</b>
             
             <h3>Translated files</h3>
             <p>
             
             As <i>pyCGNS</i> uses only GCNS/HDF5 files, using <i>CHLone</i>, the CGNS/ADF
             files are translated on the fly when you load them. The translation tool is
             <i>cgnsconvert</i> and its location should be set into the <i>Option panel</i>.
             
             A translated file is stored into a temporary directory, the <i>Info panel</i>
             shows you which is the actual original file name and the temporary filename.
             """),
    # --------------------------------------------------------------------------
    'Pattern': ('Pattern panel',
                """
                <h2>Pattern panel</h2>
                
                """),
    # --------------------------------------------------------------------------
    'Tools': ('Tools panel',
              """
              <h2>Tools panel</h2>
              
              <h1>Search</h1>
              The SIDS type can be selected as a choice in the CGNS/SIDS types or you
              can enter your own pattern for regexp search. In that later case, you have to
              press [ENTER] once you have typed down the actual search string.
              
              The Children checks are performed with a OR clause.
              If you require a child to be named 'FamilyName' for example, then the check
              is True if at least one child has this name. With the NOT clause, the check
              is True if NO child has the name 'FamilyName'.
              
              """),
    # --------------------------------------------------------------------------
    'Message': ('Message window',
                """
                <h2>About Messages</h2>
                
                Some info messages have the <i>don't show again</b> check box.
                If you set this once, you would never see the message again, unless you
                re-activate it in the <b>Option Panel</b>
                
                """),
    # --------------------------------------------------------------------------
}

# --- last line
